#
# Server support for DBus containers
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from dasbus.namespace import get_dbus_path
from dasbus.server.publishable import Publishable
from dasbus.typing import ObjPath, List

__all__ = [
    "DBusContainerError",
    "DBusContainer"
]


class DBusContainerError(Exception):
    """General exception for DBus container errors."""
    pass


class DBusContainer(object):
    """The container of DBus objects.

    A DBus container should be used to dynamically publish Publishable
    objects within the same namespace. It generates a unique DBus path
    for each object. It is able to resolve a DBus path into an object
    and an object into a DBus path.

    Example:

    .. code-block:: python

        # Create a container of tasks.
        container = DBusContainer(
            namespace=("my", "project"),
            basename="Task",
            message_bus=DBus
        )

        # Publish a task.
        path = container.to_object_path(MyTask())

        # Resolve an object path into a task.
        task = container.from_object_path(path)

    """

    def __init__(self, message_bus, namespace, basename=None):
        """Create a new container.

        :param message_bus: a message bus
        :param namespace: a sequence of names
        :param basename: a string with the base name
        """
        self._message_bus = message_bus

        if basename:
            namespace = (*namespace, basename)

        self._namespace = namespace[:-1]
        self._basename = namespace[-1]

        self._container = {}
        self._published = set()
        self._counter = 0

    def set_namespace(self, namespace):
        """Set the namespace.

        All DBus objects from the container should use the same
        namespace, so the namespace should be set up before any
        of the DBus objects are published.

        :param namespace: a sequence of names
        """
        self._namespace = namespace

    def from_object_path(self, object_path: ObjPath):
        """Convert a DBus path to a published object.

        If no published object is found for the given DBus path,
        raise DBusContainerError.

        :param object_path: a DBus path
        :return: a published object
        """
        return self._find_object(object_path)

    def to_object_path(self, obj) -> ObjPath:
        """Convert a publishable object to a DBus path.

        If no DBus path is found for the given object, publish
        the object on the container message bus with a unique
        DBus path generated from the container namespace.

        :param obj: a publishable object
        :return: a DBus path
        """
        if not isinstance(obj, Publishable):
            raise TypeError(
                "Type '{}' is not publishable.".format(type(obj).__name__)
            )

        if not self._is_object_published(obj):
            self._publish_object(obj)

        return self._find_object_path(obj)

    def from_object_path_list(self, object_paths: List[ObjPath]):
        """Convert DBus paths to published objects.

        :param object_paths: a list of DBus paths
        :return: a list of published objects
        """
        return list(map(self.from_object_path, object_paths))

    def to_object_path_list(self, objects) -> List[ObjPath]:
        """Convert publishable objects to DBus paths.

        :param objects: a list of publishable objects
        :return: a list of DBus paths
        """
        return list(map(self.to_object_path, objects))

    def _is_object_published(self, obj):
        """Is the given object published?

        :param obj: an object
        :return: True if the object is published, otherwise False
        """
        return id(obj) in self._published

    def _publish_object(self, obj: Publishable):
        """Publish the given object.

        :param obj: an object to publish
        :return: an object path
        """
        object_path = self._generate_object_path()

        self._message_bus.publish_object(
            object_path,
            obj.for_publication()
        )

        self._container[object_path] = obj
        self._published.add(id(obj))
        return object_path

    def _find_object_path(self, obj):
        """Find a DBus path of the object.

        :param obj: a published object
        :return: a DBus path
        :raise: DBusContainerError if no object path is found
        """
        for object_path, found_obj in self._container.items():
            if found_obj is obj:
                return object_path

        raise DBusContainerError(
            "No object path found."
        )

    def _find_object(self, object_path):
        """Find an object by its DBus path.

        :param object_path: a DBus path
        :return: a published object
        :raise: DBusContainerError if no object is found
        """
        if object_path in self._container:
            return self._container[object_path]

        raise DBusContainerError(
            "Unknown object path '{}'.".format(object_path)
        )

    def _generate_object_path(self):
        """Generate a unique object path.

        This method is not thread safe.

        :return: a unique object path
        """
        self._counter += 1

        return get_dbus_path(
            *self._namespace,
            self._basename,
            str(self._counter)
        )



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from collections import defaultdict
from unittest.mock import Mock, patch

from dasbus.connection import MessageBus, SystemMessageBus, \
    SessionMessageBus, AddressedMessageBus
from dasbus.constants import DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER, \
    DBUS_NAME_FLAG_ALLOW_REPLACEMENT, DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER
from dasbus.error import ErrorMapper

import gi
gi.require_version("Gio", "2.0")
from gi.repository import Gio


class MockMessageBus(MessageBus):
    """Message bus for testing."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._proxy_factory = Mock()
        self._server_factory = Mock()

    def _get_connection(self):
        return Mock()

    def publish_object(self, *args, **kwargs):  # pylint: disable=signature-differs
        return super().publish_object(
            *args, **kwargs, server_factory=self._server_factory
        )

    def get_proxy(self, *args, **kwargs):  # pylint: disable=signature-differs
        return super().get_proxy(
            *args, **kwargs, proxy_factory=self._proxy_factory
        )


class DBusConnectionTestCase(unittest.TestCase):
    """Test DBus connection."""

    def setUp(self):
        self.message_bus = MockMessageBus()
        self.error_mapper = self.message_bus._error_mapper
        self.proxy_factory = self.message_bus._proxy_factory
        self.server_factory = self.message_bus._server_factory

    def test_connection(self):
        """Test the bus connection."""
        self.assertIsNotNone(self.message_bus.connection)
        self.assertEqual(self.message_bus.connection,
                         self.message_bus.connection)
        self.assertTrue(self.message_bus.check_connection())

    def test_failing_connection(self):
        """Test the failing connection."""
        self.message_bus._get_connection = Mock(side_effect=IOError())
        with self.assertLogs(level='WARN'):
            self.assertFalse(self.message_bus.check_connection())

        self.message_bus._get_connection = Mock(return_value=None)
        self.assertFalse(self.message_bus.check_connection())

    def test_error_mapper(self):
        """Test the error mapper."""
        error_mapper = ErrorMapper()

        message_bus = MockMessageBus(error_mapper=error_mapper)
        self.assertEqual(message_bus._error_mapper, error_mapper)

        message_bus = MockMessageBus()
        self.assertNotEqual(message_bus._error_mapper, error_mapper)
        self.assertIsInstance(message_bus._error_mapper, ErrorMapper)

    def test_proxy(self):
        """Test the object proxy."""
        proxy = self.message_bus.get_proxy(
            "service.name",
            "/object/path"
        )

        self.proxy_factory.assert_called_once_with(
            self.message_bus,
            "service.name",
            "/object/path",
            error_mapper=self.error_mapper
        )

        self.assertEqual(proxy, self.proxy_factory.return_value)

    def test_interface_proxy(self):
        """Test the interface proxy."""
        proxy = self.message_bus.get_proxy(
            "service.name",
            "/object/path",
            "interface.name"
        )

        self.proxy_factory.assert_called_once_with(
            self.message_bus,
            "service.name",
            "/object/path",
            interface_name="interface.name",
            error_mapper=self.error_mapper
        )

        self.assertEqual(proxy, self.proxy_factory.return_value)

    def test_bus_proxy(self):
        """Test the bus proxy."""
        proxy = self.message_bus.proxy

        self.proxy_factory.assert_called_once_with(
            self.message_bus,
            "org.freedesktop.DBus",
            "/org/freedesktop/DBus",
            error_mapper=self.error_mapper
        )

        self.assertIsNotNone(proxy)
        self.assertEqual(proxy, self.proxy_factory.return_value)
        self.assertEqual(self.message_bus.proxy, self.message_bus.proxy)

    def test_register_service(self):
        """Test the service registration."""
        self.message_bus.proxy.RequestName.return_value = \
            DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER

        self.message_bus.register_service(
            "my.service",
            DBUS_NAME_FLAG_ALLOW_REPLACEMENT
        )

        self.message_bus.proxy.RequestName.assert_called_once_with(
            "my.service",
            DBUS_NAME_FLAG_ALLOW_REPLACEMENT
        )

        self.assertIn("my.service", self.message_bus._requested_names)
        callback = self.message_bus._registrations["my.service"]
        self.assertTrue(callable(callback))

        self.message_bus.disconnect()
        self.message_bus.proxy.ReleaseName.assert_called_once_with(
            "my.service"
        )

    def test_failed_register_service(self):
        """Test the failing service registration."""
        self.message_bus.proxy.RequestName.return_value = \
            DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER

        with self.assertRaises(ConnectionError):
            self.message_bus.register_service("my.service")

        self.message_bus.proxy.RequestName.assert_called_once_with(
            "my.service",
            DBUS_NAME_FLAG_ALLOW_REPLACEMENT
        )

        self.assertNotIn("my.service", self.message_bus._requested_names)

    def test_check_service_access(self):
        """Check the service access."""
        # The service can be accessed.
        self.message_bus.get_proxy("my.service", "/my/object")

        # The service cannot be accessed.
        self.message_bus.proxy.RequestName.return_value = \
            DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER

        self.message_bus.register_service("my.service")

        with self.assertRaises(RuntimeError) as cm:
            self.message_bus.get_proxy("my.service", "/my/object")

        self.assertEqual(
            "Can't access DBus service 'my.service' from the main thread.",
            str(cm.exception)
        )

    def test_publish_object(self):
        """Test the object publishing."""
        obj = Mock()
        self.message_bus.publish_object("/my/object", obj)

        self.server_factory.assert_called_once_with(
            self.message_bus,
            "/my/object",
            obj,
            error_mapper=self.error_mapper
        )

        callback = self.message_bus._registrations["/my/object"]
        self.assertTrue(callable(callback))

        self.message_bus.disconnect()
        callback.assert_called_once_with()

    def test_disconnect(self):
        """Test the disconnection."""
        # Set up the connection.
        self.assertIsNotNone(self.message_bus.connection)

        # Create registrations.
        callbacks = defaultdict(Mock)

        self.message_bus._registrations = {
            "my.service.1": callbacks["my.service.1"],
            "my.service.2": callbacks["my.service.2"],
            "/my/object/1": callbacks["/my/object/1"],
            "/my/object/2": callbacks["/my/object/2"],
        }

        self.message_bus._requested_names = {
            "my.service.1",
            "my.service.2"
        }

        # Disconnect.
        self.message_bus.disconnect()
        self.assertEqual(self.message_bus._connection, None)
        self.assertEqual(self.message_bus._registrations, {})
        self.assertEqual(self.message_bus._requested_names, set())

        for callback in callbacks.values():
            callback.assert_called_once_with()

        # Do nothing by default.
        self.message_bus.disconnect()

    def test_remove_single_object(self):
        """Test the disconnection."""
        # Set up the connection.
        self.assertIsNotNone(self.message_bus.connection)

        # Create registrations.
        callbacks = defaultdict(Mock)

        self.message_bus._registrations = {
            "my.service.1": callbacks["my.service.1"],
            "my.service.2": callbacks["my.service.2"],
            "/my/object/1": callbacks["/my/object/1"],
            "/my/object/2": callbacks["/my/object/2"],
        }

        self.message_bus._requested_names = {
            "my.service.1",
            "my.service.2"
        }

        # Disconnect.
        self.message_bus.unregister_service("my.service.1")
        self.assertEqual(len(self.message_bus._registrations), 3)

        # Disconnect.
        self.message_bus.unpublish_object("/my/object/2")
        self.assertEqual(len(self.message_bus._registrations), 2)

        callbacks["my.service.1"].assert_called_once_with()
        callbacks["/my/object/2"].assert_called_once_with()

    @patch("dasbus.connection.Gio.bus_get_sync")
    def test_system_bus(self, getter):
        """Test the system bus."""
        message_bus = SystemMessageBus()
        self.assertIsNotNone(message_bus.connection)
        getter.assert_called_once_with(
            Gio.BusType.SYSTEM,
            None
        )

    @patch("dasbus.connection.Gio.bus_get_sync")
    def test_session_bus(self, getter):
        """Test the session bus."""
        message_bus = SessionMessageBus()
        self.assertIsNotNone(message_bus.connection)
        getter.assert_called_once_with(
            Gio.BusType.SESSION,
            None
        )

    def _check_addressed_connection(self, message_bus, getter, address):
        self.assertIsNotNone(message_bus.connection)
        self.assertEqual(message_bus.address, address)
        getter.assert_called_once_with(
            address,
            (
                Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT |
                Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION
            ),
            None,
            None
        )

    @patch("dasbus.connection.Gio.DBusConnection.new_for_address_sync")
    def test_addressed_bus(self, getter):
        """Test the addressed bus."""
        message_bus = AddressedMessageBus("ADDRESS")
        self._check_addressed_connection(message_bus, getter, "ADDRESS")



#
# Send a notification to the notification server.
#
from common import NOTIFICATIONS

if __name__ == "__main__":
    # Create a proxy of the object /org/freedesktop/Notifications
    # provided by the service org.freedesktop.Notifications.
    proxy = NOTIFICATIONS.get_proxy()

    # Call the DBus method Notify.
    notification_id = proxy.Notify(
        "", 0, "face-smile", "Hello World!",
        "This notification can be ignored.",
        [], {}, 0
    )

    # Print the return value of the call.
    print("The notification {} was sent.".format(notification_id))



#
# Handle changed properties.
# Start the server, start the listener and run the client.
#
from dasbus.loop import EventLoop
from common import REGISTER


def callback(interface, changed_properties, invalid_properties):
    """The callback of the DBus signal PropertiesChanged."""
    print("Properties of {} has changed: {}".format(
        interface, changed_properties
    ))


if __name__ == "__main__":
    # Create a proxy of the object /org/example/Register
    # provided by the service org.example.Register
    proxy = REGISTER.get_proxy()

    # Connect the callback to the DBus signal PropertiesChanged.
    proxy.PropertiesChanged.connect(callback)

    # Start the event loop.
    loop = EventLoop()
    loop.run()



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#

import unittest

from typing import Set

from dasbus.typing import get_dbus_type, is_base_type, get_native, \
    get_variant, get_variant_type, Int, Int16, Int32, Int64, UInt16, UInt32, \
    UInt64, Bool, Byte, Str, Dict, List, Tuple, Variant, Double, ObjPath, \
    UnixFD, unwrap_variant, get_type_name, is_tuple_of_one, \
    get_type_arguments, VariantUnpacker, VariantUnwrapper

import gi
gi.require_version("GLib", "2.0")
from gi.repository import GLib


class DBusTypingTests(unittest.TestCase):

    def _compare(self, type_hint, expected_string):
        """Compare generated and expected types."""
        # Generate a type string.
        dbus_type = get_dbus_type(type_hint)
        self.assertEqual(dbus_type, expected_string)
        self.assertTrue(GLib.VariantType.string_is_valid(dbus_type))

        # Create a variant type from a type hint.
        variant_type = get_variant_type(type_hint)
        self.assertIsInstance(variant_type, GLib.VariantType)
        self.assertEqual(variant_type.dup_string(), expected_string)

        expected_type = GLib.VariantType.new(expected_string)
        self.assertTrue(expected_type.equal(variant_type))

        # Create a variant type from a type string.
        variant_type = get_variant_type(expected_string)
        self.assertIsInstance(variant_type, GLib.VariantType)
        self.assertTrue(expected_type.equal(variant_type))

        # Test the is_tuple_of_one function.
        expected_value = is_base_type(type_hint, Tuple) \
            and len(get_type_arguments(type_hint)) == 1

        self.assertEqual(is_tuple_of_one(type_hint), expected_value)
        self.assertEqual(is_tuple_of_one(expected_string), expected_value)

        self.assertTrue(is_tuple_of_one(Tuple[type_hint]))
        self.assertTrue(is_tuple_of_one("({})".format(expected_string)))

    def test_unknown(self):
        """Test the unknown type."""

        class UnknownType:
            pass

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(UnknownType)

        self.assertEqual(
            "Invalid DBus type 'UnknownType'.",
            str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(List[UnknownType])

        self.assertEqual(
            "Invalid DBus type 'UnknownType'.",
            str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(Tuple[Int, Str, UnknownType])

        self.assertEqual(
            "Invalid DBus type 'UnknownType'.",
            str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(Dict[Int, UnknownType])

        self.assertEqual(
            "Invalid DBus type 'UnknownType'.",
            str(cm.exception)
        )

    def test_invalid(self):
        """Test the invalid types."""
        msg = "Invalid DBus type of dictionary key: '{}'"

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(Dict[List[Bool], Bool])

        self.assertEqual(
            msg.format(get_type_name(List[Bool])),
            str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(Dict[Variant, Int])

        self.assertEqual(
            msg.format("Variant"),
            str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(Tuple[Int, Double, Dict[Tuple[Int, Int], Bool]])

        self.assertEqual(
            msg.format(get_type_name(Tuple[Int, Int])),
            str(cm.exception)
        )

        msg = "Invalid DBus type '{}'."

        with self.assertRaises(TypeError) as cm:
            get_dbus_type(Set[Int])

        self.assertEqual(
            msg.format(get_type_name(Set[Int])),
            str(cm.exception),
        )

    def test_simple(self):
        """Test simple types."""
        self._compare(int, "i")
        self._compare(bool, "b")
        self._compare(float, "d")
        self._compare(str, "s")

    def test_basic(self):
        """Test basic types."""
        self._compare(Int, "i")
        self._compare(Bool, "b")
        self._compare(Double, "d")
        self._compare(Str, "s")
        self._compare(ObjPath, "o")
        self._compare(UnixFD, "h")
        self._compare(Variant, "v")

    def test_int(self):
        """Test integer types."""
        self._compare(Byte, "y")
        self._compare(Int16, "n")
        self._compare(UInt16, "q")
        self._compare(Int32, "i")
        self._compare(UInt32, "u")
        self._compare(Int64, "x")
        self._compare(UInt64, "t")

    def test_container(self):
        """Test container types."""
        self._compare(Tuple[Bool], "(b)")
        self._compare(Tuple[Int, Str], "(is)")
        self._compare(Tuple[UnixFD, Variant, Double], "(hvd)")

        self._compare(List[Int], "ai")
        self._compare(List[Bool], "ab")
        self._compare(List[UnixFD], "ah")
        self._compare(List[ObjPath], "ao")

        self._compare(Dict[Str, Int], "a{si}")
        self._compare(Dict[Int, Bool], "a{ib}")

    def test_alias(self):
        """Test type aliases."""
        AliasType = List[Double]
        self._compare(Dict[Str, AliasType], "a{sad}")

    def test_depth(self):
        """Test difficult type structures."""
        self._compare(Tuple[Int, Tuple[Str, Str]], "(i(ss))")
        self._compare(Tuple[Tuple[Tuple[Int]]], "(((i)))")
        self._compare(Tuple[Bool, Tuple[Tuple[Int], Str]], "(b((i)s))")

        self._compare(List[List[List[Int]]], "aaai")
        self._compare(List[Tuple[Dict[Str, Int]]], "a(a{si})")
        self._compare(List[Dict[Str, Tuple[UnixFD, Variant]]], "aa{s(hv)}")

        self._compare(Dict[Str, List[Bool]], "a{sab}")
        self._compare(Dict[Str, Tuple[Int, Int, Double]], "a{s(iid)}")
        self._compare(Dict[Str, Tuple[Int, Int, Dict[Int, Str]]],
                      "a{s(iia{is})}")

    def test_base_type(self):
        """Test the base type checks."""
        self.assertEqual(is_base_type(Int, Int), True)
        self.assertEqual(is_base_type(UInt16, UInt16), True)
        self.assertEqual(is_base_type(Variant, Variant), True)

        self.assertEqual(is_base_type(Int, Bool), False)
        self.assertEqual(is_base_type(Bool, List), False)
        self.assertEqual(is_base_type(UInt16, Dict), False)
        self.assertEqual(is_base_type(UInt16, Int), False)
        self.assertEqual(is_base_type(Variant, Tuple), False)

        self.assertEqual(is_base_type(List[Int], List), True)
        self.assertEqual(is_base_type(List[Bool], List), True)
        self.assertEqual(is_base_type(List[Variant], List), True)

        self.assertEqual(is_base_type(Tuple[Int], Tuple), True)
        self.assertEqual(is_base_type(Tuple[Bool], Tuple), True)
        self.assertEqual(is_base_type(Tuple[Variant], Tuple), True)

        self.assertEqual(is_base_type(Dict[Str, Int], Dict), True)
        self.assertEqual(is_base_type(Dict[Str, Bool], Dict), True)
        self.assertEqual(is_base_type(Dict[Str, Variant], Dict), True)

        self.assertEqual(is_base_type(List[Int], Tuple), False)
        self.assertEqual(is_base_type(Tuple[Bool], Dict), False)
        self.assertEqual(is_base_type(Dict[Str, Variant], List), False)

    def test_base_class(self):
        """Test the base class checks."""
        class Data(object):
            pass

        class DataA(Data):
            pass

        class DataB(Data):
            pass

        self.assertEqual(is_base_type(Data, Data), True)
        self.assertEqual(is_base_type(DataA, Data), True)
        self.assertEqual(is_base_type(DataB, Data), True)

        self.assertEqual(is_base_type(Data, DataA), False)
        self.assertEqual(is_base_type(Data, DataB), False)
        self.assertEqual(is_base_type(DataA, DataB), False)
        self.assertEqual(is_base_type(DataB, DataA), False)

        self.assertEqual(is_base_type(List[Data], List), True)
        self.assertEqual(is_base_type(Tuple[DataA], Tuple), True)
        self.assertEqual(is_base_type(Dict[Str, DataB], Dict), True)


class DBusTypingVariantTests(unittest.TestCase):

    def _test_variant(self, type_hint, expected_string, value,
                      native_value=None):
        """Create a variant."""
        # The value is native by default.
        if native_value is None:
            native_value = value

        # Create a variant from a type hint.
        v1 = get_variant(type_hint, value)
        self.assertTrue(isinstance(v1, Variant))
        self.assertEqual(v1.format_string, expected_string)

        # Check unpacking.
        self.assertEqual(v1.unpack(), native_value)
        self.assertEqual(VariantUnpacker.apply(v1), native_value)

        # Check unwrapping.
        self.assertEqual(unwrap_variant(v1), value)
        self.assertEqual(VariantUnwrapper.apply(v1), value)

        # Create a variant from a type string.
        v2 = Variant(expected_string, value)
        self.assertTrue(v2.equal(v1))

        self.assertEqual(get_native(v1), native_value)
        self.assertEqual(get_native(v1), get_native(v2))
        self.assertEqual(get_native(value), native_value)

        # Create a variant from a type string.
        v3 = get_variant(expected_string, value)
        self.assertTrue(isinstance(v3, Variant))
        self.assertTrue(v2.equal(v3))

    def test_variant_invalid(self):
        """Test invalid variants."""

        class UnknownType:
            pass

        with self.assertRaises(TypeError) as cm:
            get_variant(UnknownType, 1)

        self.assertEqual(
            "Invalid DBus type 'UnknownType'.",
            str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            get_variant(Int, None)

        self.assertEqual(
            "Invalid DBus value 'None'.",
            str(cm.exception)
        )

        with self.assertRaises(TypeError):
            get_variant(List[Int], True)

    def test_variant_basic(self):
        """Test variants with basic types."""
        self._test_variant(Int, "i", 1)
        self._test_variant(Bool, "b", True)
        self._test_variant(Double, "d", 1.0)
        self._test_variant(Str, "s", "Hi!")
        self._test_variant(ObjPath, "o", "/org/something")

    def test_variant_int(self):
        """Test variants with integer types."""
        self._test_variant(Int16, "n", 2)
        self._test_variant(UInt16, "q", 3)
        self._test_variant(Int32, "i", 4)
        self._test_variant(UInt32, "u", 5)
        self._test_variant(Int64, "x", 6)
        self._test_variant(UInt64, "t", 7)

    def test_variant_container(self):
        """Test variants with container types."""
        self._test_variant(Tuple[Bool], "(b)", (False,))
        self._test_variant(Tuple[Int, Str], "(is)", (0, "zero"))

        self._test_variant(List[Int], "ai", [1, 2, 3])
        self._test_variant(List[Bool], "ab", [True, False, True])

        self._test_variant(Dict[Str, Int], "a{si}", {"a": 1, "b": 2})
        self._test_variant(Dict[Int, Bool], "a{ib}", {1: True, 2: False})

    def test_variant_variant(self):
        """Test variants with variant types."""
        variant = get_variant("i", 1)
        self._test_variant(Variant, "v", variant, 1)

        variant = get_variant("v", get_variant("i", 1))
        self._test_variant(Variant, "v", variant, 1)

        variant = get_variant("(b)", (False,))
        self._test_variant(Variant, "v", variant, (False,))

        variant = get_variant("(v)", (get_variant("b", False),))
        self._test_variant(Variant, "v", variant, (False,))

        variant = get_variant("ai", [1, 2, 3])
        self._test_variant(Variant, "v", variant, [1, 2, 3])

        variant = get_variant("av", [
            get_variant("i", 1),
            get_variant("i", 2),
            get_variant("i", 3)
        ])
        self._test_variant(Variant, "v", variant, [1, 2, 3])

    def test_variant_alias(self):
        """Test variants with type aliases."""
        AliasType = List[Double]
        self._test_variant(Dict[Str, AliasType], "a{sad}", {
            "test": [1.1, 2.2]
        })

    def _test_native(self, variants, values):
        """Test native values of variants."""
        for variant, value in zip(variants, values):
            self.assertEqual(get_native(variant), value)

        self.assertEqual(get_native(tuple(variants)), tuple(values))
        self.assertEqual(get_native(list(variants)), list(values))
        self.assertEqual(get_native(dict(enumerate(variants))),
                         dict(enumerate(values)))

        variant = get_variant(
            Tuple[Variant, Variant, Variant, Variant],
            tuple(variants)
        )
        self.assertEqual(unwrap_variant(variant), tuple(variants))

        variant = get_variant(
            List[Variant],
            list(variants)
        )
        self.assertEqual(unwrap_variant(variant), list(variants))

        variant = get_variant(
            Dict[Int, Variant],
            dict(enumerate(variants))
        )
        self.assertEqual(unwrap_variant(variant), dict(enumerate(variants)))

    def test_basic_native(self):
        """Test get_native with basic variants."""
        self._test_native(
            [
                get_variant(Double, 1.2),
                get_variant(List[Int], [0, -1]),
                get_variant(Tuple[Bool, Bool], (True, False)),
                get_variant(Dict[Str, Int], {"key": 0}),
            ],
            [
                1.2,
                [0, -1],
                (True, False),
                {"key": 0}
            ]
        )

    def test_complex_native(self):
        """Test get_native with complex variants."""
        self._test_native(
            [
                get_variant(Variant, get_variant(Double, 1.2)),
                get_variant(List[Variant], [
                    get_variant(Int, 0),
                    get_variant(Int, -1)
                ]),
                get_variant(Tuple[Variant, Bool], (
                    get_variant(Bool, True),
                    False
                )),
                get_variant(Dict[Str, Variant], {
                    "key": get_variant(Int, 0)
                })
            ],
            [
                1.2,
                [0, -1],
                (True, False),
                {"key": 0}
            ]
        )



#
# The common definitions
#
from dasbus.connection import SessionMessageBus
from dasbus.identifier import DBusServiceIdentifier, DBusInterfaceIdentifier
from dasbus.server.container import DBusContainer

# Define the message bus.
SESSION_BUS = SessionMessageBus()

# Define namespaces.
CHAT_NAMESPACE = ("org", "example", "Chat")
ROOMS_NAMESPACE = (*CHAT_NAMESPACE, "Rooms")

# Define services and objects.
CHAT = DBusServiceIdentifier(
    namespace=CHAT_NAMESPACE,
    message_bus=SESSION_BUS
)

ROOM = DBusInterfaceIdentifier(
    namespace=CHAT_NAMESPACE,
    basename="Room"
)

# Define containers.
ROOM_CONTAINER = DBusContainer(
    namespace=ROOMS_NAMESPACE,
    message_bus=SESSION_BUS
)



#
# Client support for DBus properties
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#

__all__ = ["PropertyProxy"]


class PropertyProxy(object):
    """Proxy of a remote DBus property.

    It can be used to define instance attributes.
    """

    __slots__ = [
        "_getter",
        "_setter"
    ]

    def __init__(self, getter, setter):
        """Create a new proxy of the DBus property."""
        self._getter = getter
        self._setter = setter

    def get(self):
        """Get the value of the DBus property."""
        return self.__get__(None, None)  # pylint: disable=unnecessary-dunder-call

    def __get__(self, instance, owner):
        if instance is None and owner:
            return self

        if not self._getter:
            raise AttributeError(
                "Can't read DBus property."
            )

        return self._getter()

    def set(self, value):
        """Set the value of the DBus property."""
        return self.__set__(None, value)  # pylint: disable=unnecessary-dunder-call

    def __set__(self, instance, value):
        if not self._setter:
            raise AttributeError(
                "Can't set DBus property."
            )

        return self._setter(value)



#
# Reply to a message in the chat room.
# Start the server, start the listener and run the client.
#
from dasbus.loop import EventLoop
from common import CHAT


def callback(proxy, msg):
    """The callback of the DBus signal MessageReceived."""
    if "I am Alice!" in msg:
        proxy.SendMessage("Hello Alice, I am Bob.")


if __name__ == "__main__":
    # Create a proxy of the object /org/example/Chat
    # provided by the service org.example.Chat.
    chat_proxy = CHAT.get_proxy()

    # Find a chat room to monitor.
    object_path = chat_proxy.FindRoom("Bob's room")

    # Create a proxy of the object /org/example/Chat/Rooms/1
    # provided by the service org.example.Chat.
    room_proxy = CHAT.get_proxy(object_path)

    # Connect the callback to the DBus signal MessageReceived.
    room_proxy.MessageReceived.connect(lambda msg: callback(room_proxy, msg))

    # Start the event loop.
    loop = EventLoop()
    loop.run()



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest

from dasbus.typing import get_variant, get_native, Int, Str, List, Bool, \
    Dict, Structure, Variant
from dasbus.structure import DBusData, DBusStructureError, compare_data, \
    generate_string_from_data


class DBusStructureTestCase(unittest.TestCase):
    """Test the DBus structure support."""

    def test_empty_structure(self):
        with self.assertRaises(DBusStructureError) as cm:
            class NoData(DBusData):
                pass

            NoData()

        self.assertEqual(str(cm.exception), "No fields found.")

    def test_readonly_structure(self):
        with self.assertRaises(DBusStructureError) as cm:
            class ReadOnlyData(DBusData):
                @property
                def x(self) -> Int:
                    return 1

            ReadOnlyData()

        self.assertEqual(str(cm.exception), "Field 'x' cannot be set.")

    def test_writeonly_structure(self):
        with self.assertRaises(DBusStructureError) as cm:
            class WriteOnlyData(DBusData):
                def __init__(self):
                    self._x = 0

                def set_x(self, x):
                    self._x = x

                x = property(None, set_x)

            WriteOnlyData()

        self.assertEqual(str(cm.exception), "Field 'x' cannot be get.")

    def test_no_type_structure(self):
        with self.assertRaises(DBusStructureError) as cm:
            class NoTypeData(DBusData):
                def __init__(self):
                    self._x = 0

                @property
                def x(self):
                    return self._x

                @x.setter
                def x(self, x):
                    self._x = x

            NoTypeData()

        self.assertEqual(str(cm.exception), "Field 'x' has unknown type.")

    class SkipData(DBusData):

        class_attribute = 1

        def __init__(self):
            self._x = 0
            self._y = 1

        @property
        def x(self) -> Int:
            return self._x

        @property
        def _private_property(self):
            return 1

        @x.setter
        def x(self, x):
            self._x = x

        def method(self):
            pass

    def test_skip_members(self):
        data = self.SkipData()

        structure = self.SkipData.to_structure(data)
        self.assertEqual(structure, {
            'x': get_variant(Int, 0)
        })

        data = self.SkipData.from_structure({
            'x': get_variant(Int, 10)
        })

        self.assertEqual(data.x, 10)

    class SimpleData(DBusData):

        def __init__(self):
            self._x = 0

        @property
        def x(self) -> Int:
            return self._x

        @x.setter
        def x(self, x):
            self._x = x

    def test_get_simple_structure(self):
        data = self.SimpleData()
        self.assertEqual(data.x, 0)

        structure = self.SimpleData.to_structure(data)
        self.assertEqual(structure, {'x': get_variant(Int, 0)})

        data.x = 10
        self.assertEqual(data.x, 10)

        structure = self.SimpleData.to_structure(data)
        self.assertEqual(structure, {'x': get_variant(Int, 10)})

    def test_get_simple_structure_list(self):
        d1 = self.SimpleData()
        d1.x = 1

        d2 = self.SimpleData()
        d2.x = 2

        d3 = self.SimpleData()
        d3.x = 3

        structures = self.SimpleData.to_structure_list([d1, d2, d3])

        self.assertEqual(structures, [
            {'x': get_variant(Int, 1)},
            {'x': get_variant(Int, 2)},
            {'x': get_variant(Int, 3)}
        ])

    def test_apply_simple_structure(self):
        data = self.SimpleData()
        self.assertEqual(data.x, 0)

        structure = {'x': get_variant(Int, 10)}
        data = self.SimpleData.from_structure(structure)

        self.assertEqual(data.x, 10)

    def test_apply_simple_invalid_structure(self):
        with self.assertRaises(DBusStructureError) as cm:
            self.SimpleData.from_structure({'y': get_variant(Int, 10)})

        self.assertEqual(str(cm.exception), "Field 'y' doesn't exist.")

    def test_apply_simple_structure_list(self):
        s1 = {'x': get_variant(Int, 1)}
        s2 = {'x': get_variant(Int, 2)}
        s3 = {'x': get_variant(Int, 3)}

        data = self.SimpleData.from_structure_list([s1, s2, s3])

        self.assertEqual(len(data), 3)
        self.assertEqual(data[0].x, 1)
        self.assertEqual(data[1].x, 2)
        self.assertEqual(data[2].x, 3)

    def test_compare_simple_structure(self):
        data = self.SimpleData()

        self.assertTrue(compare_data(data, data))
        self.assertTrue(compare_data(data, self.SimpleData()))

        self.assertFalse(compare_data(data, None))
        self.assertFalse(compare_data(None, data))

        self.assertTrue(compare_data(
            self.SimpleData.from_structure({'x': get_variant(Int, 10)}),
            self.SimpleData.from_structure({'x': get_variant(Int, 10)})
        ))

        self.assertFalse(compare_data(
            self.SimpleData.from_structure({'x': get_variant(Int, 10)}),
            self.SimpleData.from_structure({'x': get_variant(Int, 9)})
        ))

    class OtherData(DBusData):

        def __init__(self):
            self._x = 0

        @property
        def x(self) -> Int:
            return self._x

        @x.setter
        def x(self, x):
            self._x = x

    def test_get_structure_from_invalid_type(self):
        data = self.OtherData()

        with self.assertRaises(TypeError) as cm:
            self.SimpleData.to_structure(data)

        self.assertEqual(str(cm.exception), "Invalid type 'OtherData'.")

    def test_apply_structure_with_invalid_type(self):
        structure = ["x"]

        with self.assertRaises(TypeError) as cm:
            self.SimpleData.from_structure(structure)

        self.assertEqual(str(cm.exception), "Invalid type 'list'.")

    def test_apply_structure_with_invalid_type_variant(self):
        structure = get_variant(List[Structure], [{'x': get_variant(Int, 10)}])

        with self.assertRaises(TypeError) as cm:
            self.SimpleData.from_structure_list(structure)

        self.assertEqual(str(cm.exception), "Invalid type 'Variant'.")

    class ComplicatedData(DBusData):

        def __init__(self):
            self._very_long_property_name = ""
            self._bool_list = []
            self._dictionary = {}

        @property
        def dictionary(self) -> Dict[Int, Str]:
            return self._dictionary

        @dictionary.setter
        def dictionary(self, value):
            self._dictionary = value

        @property
        def bool_list(self) -> List[Bool]:
            return self._bool_list

        @bool_list.setter
        def bool_list(self, value):
            self._bool_list = value

        @property
        def very_long_property_name(self) -> Str:
            return self._very_long_property_name

        @very_long_property_name.setter
        def very_long_property_name(self, value):
            self._very_long_property_name = value

    def test_get_complicated_structure(self):
        data = self.ComplicatedData()
        data.dictionary = {1: "1", 2: "2"}
        data.bool_list = [True, False, False]
        data.very_long_property_name = "My String Value"

        self.assertEqual(
            {
                'dictionary': get_variant(
                    Dict[Int, Str], {1: "1", 2: "2"}
                ),
                'bool-list': get_variant(
                    List[Bool], [True, False, False]
                ),
                'very-long-property-name': get_variant(
                    Str, "My String Value"
                )
            },
            self.ComplicatedData.to_structure(data)
        )

    def test_apply_complicated_structure(self):
        data = self.ComplicatedData.from_structure(
            {
                'dictionary': get_variant(
                    Dict[Int, Str], {1: "1", 2: "2"}
                ),
                'bool-list': get_variant(
                    List[Bool], [True, False, False]
                ),
                'very-long-property-name': get_variant(
                    Str, "My String Value"
                )
            }
        )

        self.assertEqual(data.dictionary, {1: "1", 2: "2"})
        self.assertEqual(data.bool_list, [True, False, False])
        self.assertEqual(data.very_long_property_name, "My String Value")

    def test_compare_complicated_structure(self):
        self.assertTrue(compare_data(
            self.ComplicatedData(),
            self.ComplicatedData(),
        ))

        self.assertFalse(compare_data(
            self.ComplicatedData(),
            self.SimpleData()
        ))

        self.assertFalse(compare_data(
            self.SimpleData(),
            self.ComplicatedData()
        ))

        self.assertTrue(compare_data(
            self.ComplicatedData.from_structure(
                {
                    'dictionary': get_variant(
                        Dict[Int, Str], {1: "1", 2: "2"}
                    ),
                    'bool-list': get_variant(
                        List[Bool], [True, False, False]
                    ),
                    'very-long-property-name': get_variant(
                        Str, "My String Value"
                    )
                }
            ),
            self.ComplicatedData.from_structure(
                {
                    'dictionary': get_variant(
                        Dict[Int, Str], {1: "1", 2: "2"}
                    ),
                    'bool-list': get_variant(
                        List[Bool], [True, False, False]
                    ),
                    'very-long-property-name': get_variant(
                        Str, "My String Value"
                    )
                }
            )
        ))

        self.assertFalse(compare_data(
            self.ComplicatedData.from_structure(
                {
                    'dictionary': get_variant(
                        Dict[Int, Str], {1: "1", 2: "2"}
                    ),
                    'bool-list': get_variant(
                        List[Bool], [True, False, False]
                    ),
                    'very-long-property-name': get_variant(
                        Str, "My String Value"
                    )
                }
            ),
            self.ComplicatedData.from_structure(
                {
                    'dictionary': get_variant(
                        Dict[Int, Str], {1: "1", 2: "2"}
                    ),
                    'bool-list': get_variant(
                        List[Bool], [True, False, True]
                    ),
                    'very-long-property-name': get_variant(
                        Str, "My String Value"
                    )
                }
            )
        ))

    def test_get_native_complicated_structure(self):
        data = self.ComplicatedData.from_structure({
            'dictionary': get_variant(
                Dict[Int, Str], {1: "1", 2: "2"}
            ),
            'bool-list': get_variant(
                List[Bool], [True, False, False]
            ),
            'very-long-property-name': get_variant(
                Str, "My String Value"
            )
        })

        structure = self.ComplicatedData.to_structure(
            data
        )

        dictionary = {
            'dictionary': {1: "1", 2: "2"},
            'bool-list': [True, False, False],
            'very-long-property-name': "My String Value"
        }

        self.assertEqual(get_native(structure), dictionary)
        self.assertEqual(get_native(dictionary), dictionary)

    class StringData(DBusData):

        def __init__(self):
            self._a = 1
            self._b = ""
            self._c = []
            self._d = []

        @property
        def a(self) -> Int:
            return self._a

        @a.setter
        def a(self, value):
            self._a = value

        @property
        def b(self) -> Str:
            return self._b

        @b.setter
        def b(self, value):
            self._b = value

        @property
        def c(self) -> List[Bool]:
            return self._c

        @c.setter
        def c(self, value):
            self._c = value

    def test_string_representation(self):
        data = self.StringData()

        expected = "StringData(a=1, b='', c=[])"
        self.assertEqual(expected, repr(data))
        self.assertEqual(expected, str(data))

        data.a = 123
        data.b = "HELLO"
        data.c = [True, False]

        expected = "StringData(a=123, b='HELLO', c=[True, False])"
        self.assertEqual(expected, repr(data))
        self.assertEqual(expected, str(data))

    class AdvancedStringData(DBusData):

        def __init__(self):
            self._a = ""
            self._b = ""
            self._c = ""

        @property
        def a(self) -> Str:
            return self._a

        @a.setter
        def a(self, value):
            self._a = value

        @property
        def b(self) -> Str:
            return self._b

        @b.setter
        def b(self, value):
            self._b = value

        @property
        def c(self) -> Str:
            return self._c

        @c.setter
        def c(self, value):
            self._c = value

        def __repr__(self):
            return generate_string_from_data(
                obj=self,
                skip=["b"],
                add={"b_is_set": bool(self.b)}
            )

    def test_advanced_string_representation(self):
        data = self.AdvancedStringData()

        expected = "AdvancedStringData(a='', b_is_set=False, c='')"
        self.assertEqual(expected, repr(data))
        self.assertEqual(expected, str(data))

        data.a = "A"
        data.b = "B"
        data.c = "C"

        expected = "AdvancedStringData(a='A', b_is_set=True, c='C')"
        self.assertEqual(expected, repr(data))
        self.assertEqual(expected, str(data))

    def test_generate_string_from_invalid_type(self):
        with self.assertRaises(DBusStructureError) as cm:
            generate_string_from_data({"x": 1})

        self.assertEqual(
            str(cm.exception),
            "Fields are not defined at '__dbus_fields__'."
        )

    def test_nested_structure(self):
        class SimpleData(DBusData):

            def __init__(self):
                self._x = 0

            @property
            def x(self) -> Int:
                return self._x

            @x.setter
            def x(self, value):
                self._x = value

        class SecretData(DBusData):

            def __init__(self):
                self._y = ""

            @property
            def y(self) -> Str:
                return self._y

            @y.setter
            def y(self, value):
                self._y = value

            def __repr__(self):
                return generate_string_from_data(
                    self, skip=["y"], add={"y_set": bool(self.y)}
                )

        class NestedData(DBusData):

            def __init__(self):
                self._attr = SimpleData()
                self._secret = SecretData()
                self._list = []

            @property
            def attr(self) -> SimpleData:
                return self._attr

            @attr.setter
            def attr(self, value):
                self._attr = value

            @property
            def secret(self) -> SecretData:
                return self._secret

            @secret.setter
            def secret(self, value):
                self._secret = value

            @property
            def list(self) -> List[SimpleData]:
                return self._list

            @list.setter
            def list(self, value):
                self._list = value

        data = NestedData()
        expected = \
            "NestedData(" \
            "attr=SimpleData(x=0), " \
            "list=[], " \
            "secret=SecretData(y_set=False))"

        self.assertEqual(str(data), expected)
        self.assertEqual(repr(data), expected)

        data.attr.x = -1
        data.secret.y = "SECRET"

        for x in range(2):
            item = SimpleData()
            item.x = x
            data.list.append(item)

        expected = \
            "NestedData(" \
            "attr=SimpleData(x=-1), " \
            "list=[SimpleData(x=0), SimpleData(x=1)], " \
            "secret=SecretData(y_set=True))"

        self.assertEqual(str(data), expected)
        self.assertEqual(repr(data), expected)

        self.assertEqual(NestedData.to_structure(data), {
            'attr': get_variant(Structure, {
                'x': get_variant(Int, -1)
            }),
            'secret': get_variant(Structure, {
                'y': get_variant(Str, "SECRET")
            }),
            'list': get_variant(List[Structure], [
                {'x': get_variant(Int, 0)},
                {'x': get_variant(Int, 1)}
            ])
        })

        dictionary = {
            'attr': {'x': 10},
            'secret': {'y': "SECRET"},
            'list': [{'x': 200}, {'x': 300}]
        }

        structure = {
            'attr': get_variant(Dict[Str, Variant], {
                'x': get_variant(Int, 10)
            }),
            'secret': get_variant(Dict[Str, Variant], {
                'y': get_variant(Str, "SECRET")
            }),
            'list': get_variant(List[Dict[Str, Variant]], [
                {'x': get_variant(Int, 200)},
                {'x': get_variant(Int, 300)}
            ])
        }

        data = NestedData.from_structure(structure)
        self.assertEqual(data.attr.x, 10)
        self.assertEqual(data.secret.y, "SECRET")
        self.assertEqual(len(data.list), 2)
        self.assertEqual(data.list[0].x, 200)
        self.assertEqual(data.list[1].x, 300)

        structure = NestedData.to_structure(data)
        self.assertEqual(get_native(structure), dictionary)



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from textwrap import dedent
from unittest.mock import Mock

from dasbus.error import ErrorMapper, ErrorRule
from dasbus.server.handler import ServerObjectHandler, GLibServer
from dasbus.server.interface import accepts_additional_arguments
from dasbus.signal import Signal
from dasbus.specification import DBusSpecificationError
from dasbus.typing import get_variant


class MethodFailedException(Exception):
    """The method has failed."""
    pass


class DBusServerTestCase(unittest.TestCase):
    """Test DBus server support."""

    NO_PARAMETERS = get_variant("()", tuple())

    def setUp(self):
        self.message_bus = Mock()
        self.connection = self.message_bus.connection
        self.error_mapper = ErrorMapper()
        self.object = None
        self.object_path = "/my/path"
        self.handler = None

    def _publish_object(self, xml="<node />"):
        """Publish a mocked object."""
        self.object = Mock(__dbus_xml__=dedent(xml))

        # Raise AttributeError for default methods.
        del self.object.Get
        del self.object.Set
        del self.object.GetAll

        # Create object signals.
        self.object.Signal1 = Signal()
        self.object.Signal2 = Signal()

        # Create default object signals.
        self.object.PropertiesChanged = Signal()

        self.handler = ServerObjectHandler(
            self.message_bus,
            self.object_path,
            self.object,
            error_mapper=self.error_mapper
        )
        self.handler.connect_object()

    def _call_method(self, interface, method, parameters=NO_PARAMETERS,
                     reply=None):
        invocation = Mock()
        invocation.get_sender.return_value = ":1.0"

        GLibServer._object_callback(
            self.connection, Mock(), self.object_path, interface, method,
            parameters, invocation, (self.handler._method_callback, ())
        )

        invocation.return_dbus_error.assert_not_called()
        invocation.return_value.assert_called_once_with(reply)

    def _call_method_with_error(self, interface, method,
                                parameters=NO_PARAMETERS,
                                error_name=None,
                                error_message=None):
        invocation = Mock()

        with self.assertLogs(level='WARN'):
            self.handler._method_callback(
                invocation,
                interface,
                method,
                parameters
            )

        invocation.return_dbus_error.assert_called_once()
        invocation.return_value.assert_not_called()

        (name, msg), kwargs = invocation.return_dbus_error.call_args

        self.assertEqual(kwargs, {})
        self.assertEqual(name, error_name, "Unexpected error name.")

        if error_message is not None:
            self.assertEqual(msg, error_message, "Unexpected error message.")

    def test_register(self):
        """Test the object registration."""
        with self.assertRaises(DBusSpecificationError) as cm:
            self._publish_object("<node />")

        self.assertEqual(
            "No DBus interfaces for registration.",
            str(cm.exception)
        )

        self._publish_object("""
        <node>
            <interface name="Interface" />
        </node>
        """)
        self.message_bus.connection.register_object.assert_called()

        self.handler.disconnect_object()
        self.message_bus.connection.unregister_object.assert_called()

    def test_method(self):
        """Test the method publishing."""
        self._publish_object("""
        <node>
            <interface name="Interface">
                <method name="Method1"/>
                <method name="Method2">
                    <arg direction="in" name="x" type="i"/>
                </method>
                <method name="Method3">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <method name="Method4">
                    <arg direction="in" name="x" type="ad"/>
                    <arg direction="in" name="y" type="o"/>
                    <arg direction="out" name="return" type="(ib)"/>
                </method>
                <method name="Method5">
                    <arg direction="out" name="return" type="i"/>
                    <arg direction="out" name="return" type="b"/>
                </method>
            </interface>
        </node>
        """)

        self.object.Method2.return_value = None
        self._call_method(
            "Interface",
            "Method2",
            parameters=get_variant("(i)", (1, ))
        )
        self.object.Method2.assert_called_once_with(1)

        self.object.Method1.return_value = None
        self._call_method("Interface", "Method1")
        self.object.Method1.assert_called_once_with()

        self.object.Method3.return_value = 0
        self._call_method(
            "Interface",
            "Method3",
            reply=get_variant("(i)", (0, ))
        )
        self.object.Method3.assert_called_once_with()

        self.object.Method4.return_value = (1, True)
        self._call_method(
            "Interface",
            "Method4",
            parameters=get_variant("(ado)", ([1.2, 2.3], "/my/path")),
            reply=get_variant("((ib))", ((1, True), ))
        )
        self.object.Method4.assert_called_once_with([1.2, 2.3], "/my/path")

        self.object.Method5.return_value = (1, True)
        self._call_method(
            "Interface",
            "Method5",
            reply=get_variant("(ib)", (1, True))
        )
        self.object.Method5.assert_called_once_with()

        self._call_method_with_error(
            "Interface",
            "MethodInvalid",
            error_name="not.known.Error.DBusSpecificationError",
            error_message="DBus specification has no member "
                          "'Interface.MethodInvalid'."
        )

        self.error_mapper.add_rule(ErrorRule(
            exception_type=MethodFailedException,
            error_name="MethodFailed"
        ))

        self.object.Method1.side_effect = MethodFailedException(
            "The method has failed."
        )
        self._call_method_with_error(
            "Interface",
            "Method1",
            error_name="MethodFailed",
            error_message="The method has failed."
        )

    def test_invalid_method_result(self):
        """Test a method with an invalid result."""
        self._publish_object("""
        <node>
            <interface name="Interface">
                <method name="Method">
                    <arg direction="out" name="return" type="t"/>
                </method>
            </interface>
        </node>
        """)

        self.object.Method.return_value = -1
        self._call_method_with_error(
            "Interface",
            "Method",
            error_name="not.known.Error.OverflowError"
        )

    def test_property(self):
        """Test the property publishing."""
        self._publish_object("""
        <node>
            <interface name="Interface">
                <property name="Property1" type="i" access="readwrite" />
                <property name="Property2" type="s" access="read" />
                <property name="Property3" type="b" access="write" />
            </interface>
        </node>
        """)

        self.object.Property1 = 0
        self._call_method(
            "org.freedesktop.DBus.Properties", "Get",
            parameters=get_variant("(ss)", (
                "Interface",
                "Property1"
            )),
            reply=get_variant("(v)", (get_variant("i", 0), ))
        )

        self._call_method(
            "org.freedesktop.DBus.Properties", "Set",
            parameters=get_variant("(ssv)", (
                "Interface",
                "Property1",
                get_variant("i", 1)
            )),
        )
        self.assertEqual(self.object.Property1, 1)

        self.object.Property2 = "Hello"
        self._call_method(
            "org.freedesktop.DBus.Properties", "Get",
            parameters=get_variant("(ss)", (
                "Interface",
                "Property2"
            )),
            reply=get_variant("(v)", (get_variant("s", "Hello"), ))
        )
        self._call_method_with_error(
            "org.freedesktop.DBus.Properties", "Set",
            parameters=get_variant("(ssv)", (
                "Interface",
                "Property2",
                get_variant("s", "World")
            )),
            error_name="not.known.Error.AttributeError",
            error_message="The property Interface.Property2 "
                          "is not writable."
        )
        self.assertEqual(self.object.Property2, "Hello")

        self.object.Property3 = True
        self._call_method_with_error(
            "org.freedesktop.DBus.Properties", "Get",
            parameters=get_variant("(ss)", (
                "Interface",
                "Property3"
            )),
            error_name="not.known.Error.AttributeError",
            error_message="The property Interface.Property3 "
                          "is not readable."
        )
        self._call_method(
            "org.freedesktop.DBus.Properties", "Set",
            parameters=get_variant("(ssv)", (
                "Interface",
                "Property3",
                get_variant("b", False)
            )),
        )
        self.assertEqual(self.object.Property3, False)

        self._call_method(
            "org.freedesktop.DBus.Properties", "GetAll",
            parameters=get_variant("(s)", ("Interface", )),
            reply=get_variant("(a{sv})", ({
                "Property1": get_variant("i", 1),
                "Property2": get_variant("s", "Hello")
            }, ))
        )

        self.object.PropertiesChanged(
            "Interface",
            {"Property1": get_variant("i", 1)},
            ["Property2"]
        )
        self.message_bus.connection.emit_signal.assert_called_once_with(
            None,
            self.object_path,
            "org.freedesktop.DBus.Properties",
            "PropertiesChanged",
            get_variant("(sa{sv}as)", (
                "Interface",
                {"Property1": get_variant("i", 1)},
                ["Property2"]
            ))
        )

    def test_signal(self):
        """Test the signal publishing."""
        self._publish_object("""
        <node>
            <interface name="Interface">
                <signal name="Signal1" />
                <signal name="Signal2">
                    <arg direction="out" name="x" type="i"/>
                    <arg direction="out" name="y" type="s"/>
                </signal>
            </interface>
        </node>
        """)
        self.object.Signal1()
        self.message_bus.connection.emit_signal.assert_called_once_with(
            None,
            self.object_path,
            "Interface",
            "Signal1",
            None
        )

        self.message_bus.connection.emit_signal.reset_mock()

        self.object.Signal2(1, "Test")
        self.message_bus.connection.emit_signal.assert_called_once_with(
            None,
            self.object_path,
            "Interface",
            "Signal2",
            get_variant("(is)", (1, "Test"))
        )

    def test_call_info(self):
        self._publish_object("""
        <node>
            <interface name="Interface">
                <method name="Method1"/>
                <method name="Method2">
                    <arg direction="in" name="x" type="i"/>
                </method>
            </interface>
        </node>
        """)

        accepts_additional_arguments(self.object.Method1)
        self._call_method("Interface", "Method1")
        self.object.Method1.assert_called_once_with(
            call_info={"sender": ":1.0"}
        )

        accepts_additional_arguments(self.object.Method2)
        self._call_method(
            "Interface",
            "Method2",
            parameters=get_variant("(i)", (1, ))
        )
        self.object.Method2.assert_called_once_with(
            1, call_info={"sender": ":1.0"}
        )



#
# Handle a closed notification.
# Start the listener, run the client and close a notification.
#
from dasbus.loop import EventLoop
from common import NOTIFICATIONS


def callback(notification_id, reason):
    """The callback of the DBus signal NotificationClosed."""
    print("The notification {} was closed.".format(notification_id))


if __name__ == "__main__":
    # Create a proxy of the object /org/freedesktop/Notifications
    # provided by the service org.freedesktop.Notifications.
    proxy = NOTIFICATIONS.get_proxy()

    # Connect the callback to the DBus signal NotificationClosed.
    proxy.NotificationClosed.connect(callback)

    # Start the event loop.
    loop = EventLoop()
    loop.run()



#
# Run the service org.example.Chat.
#
from dasbus.loop import EventLoop
from dasbus.server.interface import dbus_interface, dbus_signal
from dasbus.server.publishable import Publishable
from dasbus.server.template import InterfaceTemplate
from dasbus.signal import Signal
from dasbus.typing import Str, ObjPath
from dasbus.xml import XMLGenerator
from common import SESSION_BUS, CHAT, ROOM, ROOM_CONTAINER


@dbus_interface(ROOM.interface_name)
class RoomInterface(InterfaceTemplate):
    """The DBus interface of the chat room."""

    def connect_signals(self):
        """Connect the signals."""
        self.implementation.message_received.connect(self.MessageReceived)

    @dbus_signal
    def MessageReceived(self, msg: Str):
        """Signal that a message has been received."""
        pass

    def SendMessage(self, msg: Str):
        """Send a message to the chat room."""
        self.implementation.send_message(msg)


class Room(Publishable):
    """The implementation of the chat room."""

    def __init__(self, name):
        self._name = name
        self._message_received = Signal()

    def for_publication(self):
        """Return a DBus representation."""
        return RoomInterface(self)

    @property
    def message_received(self):
        """Signal that a message has been received."""
        return self._message_received

    def send_message(self, msg):
        """Send a message to the chat room."""
        print("{}: {}".format(self._name, msg))
        self.message_received.emit(msg)


@dbus_interface(CHAT.interface_name)
class ChatInterface(InterfaceTemplate):
    """The DBus interface of the chat service."""

    def FindRoom(self, name: Str) -> ObjPath:
        """Find or create a chat room."""
        return ROOM_CONTAINER.to_object_path(
            self.implementation.find_room(name)
        )


class Chat(Publishable):
    """The implementation of the chat."""

    def __init__(self):
        self._rooms = {}

    def for_publication(self):
        """Return a DBus representation."""
        return ChatInterface(self)

    def find_room(self, name):
        """Find or create a chat room."""
        if name not in self._rooms:
            self._rooms[name] = Room(name)

        return self._rooms[name]


if __name__ == "__main__":
    # Print the generated XML specifications.
    print(XMLGenerator.prettify_xml(ChatInterface.__dbus_xml__))
    print(XMLGenerator.prettify_xml(RoomInterface.__dbus_xml__))

    try:
        # Create the chat.
        chat = Chat()

        # Publish the chat at /org/example/Chat.
        SESSION_BUS.publish_object(CHAT.object_path, chat.for_publication())

        # Register the service name org.example.Chat.
        SESSION_BUS.register_service(CHAT.service_name)

        # Start the event loop.
        loop = EventLoop()
        loop.run()
    finally:
        # Unregister the DBus service and objects.
        SESSION_BUS.disconnect()



#
# Register the user Alice.
# Start the server and run the client twice.
#
from common import REGISTER, User, InvalidUser

if __name__ == "__main__":
    # Create a proxy of the object /org/example/Register
    # provided by the service org.example.Register
    proxy = REGISTER.get_proxy()

    # Register Alice.
    alice = User()
    alice.name = "Alice"
    alice.age = 1000

    print("Sending a DBus structure:")
    print(User.to_structure(alice))

    try:
        proxy.RegisterUser(User.to_structure(alice))
    except InvalidUser as e:
        print("Failed to register a user:", e)
        exit(1)

    # Print the registered users.
    print("Receiving DBus structures:")
    for user in proxy.Users:
        print(user)

    print("Registered users:")
    for user in User.from_structure_list(proxy.Users):
        print(user.name)



#
# Copyright (C) 2022  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import os
import tempfile
import unittest
import unittest.mock

from dasbus.connection import AddressedMessageBus
from dasbus.typing import get_variant, UnixFD, Str, Variant, Tuple, \
    List, Dict, Int, Double, Bool
from dasbus.server.interface import dbus_interface, dbus_signal
from dasbus.unix import GLibClientUnix, GLibServerUnix, acquire_fds, \
    restore_fds
from dasbus.xml import XMLGenerator

from tests.lib_dbus import run_loop, DBusSpawnedTestCase
from tests.test_dbus import DBusExampleTestCase, error_mapper

import gi
gi.require_version("Gio", "2.0")
gi.require_version("GLib", "2.0")
from gi.repository import Gio

__all__ = [
    "UnixFDSwapTests",
    "DBusSpawnedTestCase",
]


def mocked(callback):
    """Wrap the local callback in a mock object."""
    return unittest.mock.Mock(side_effect=callback)


def read_string(fd):
    """Read a value from the given file descriptor."""
    with open(os.dup(fd), "rb", closefd=True) as i:
        value = i.read().decode("utf-8")

    return value


def write_string(value):
    """Write a value to the given file descriptor."""
    with tempfile.TemporaryFile(mode="wb", buffering=0) as o:
        o.write(value.encode("utf-8"))
        o.seek(0)
        out = os.dup(o.fileno())

    return UnixFD(out)


@dbus_interface("my.testing.UnixExample")
class UnixExampleInterface(object):
    """A DBus interface with UnixFD types."""

    def __init__(self):
        self._pipes = []

    @property
    def Pipes(self) -> List[UnixFD]:
        return self._pipes

    @Pipes.setter
    def Pipes(self, pipes: List[UnixFD]):
        self._pipes = pipes

    def Hello(self, name_fd: UnixFD) -> Str:
        name = read_string(name_fd)
        return "Hello, {0}!".format(name)

    def Goodbye(self, name: Str) -> UnixFD:
        return write_string("Goodbye, {0}!".format(name))

    @dbus_signal
    def Signal(self, name: Str, name_fd: UnixFD):
        pass

    def Trigger(self, name: Str, name_fd: UnixFD):
        self.Signal(name, name_fd)


class UnixMessageBus(AddressedMessageBus):
    """A message bus connection with the UnixFD support."""

    def publish_object(self, *args, **kwargs):
        return super().publish_object(*args, **kwargs, server=GLibServerUnix)

    def get_proxy(self, *args, **kwargs):
        return super().get_proxy(*args, **kwargs, client=GLibClientUnix)


class UnixFDSwapTests(unittest.TestCase):
    """Test swapping of Unix file descriptors."""

    def setUp(self):
        """Set up the test."""
        self._r, self._w = os.pipe()

    def tearDown(self):
        """Tear down the test."""
        os.close(self._r)
        os.close(self._w)

    def _swap_fds(self, in_variant, out_variant, fds=None):
        """Swap Unix file descriptors in the input."""
        variant, fd_list = self._acquire_fds(in_variant, out_variant, fds)
        self._restore_fds(variant, fd_list, expected_variant=in_variant)

    def _acquire_fds(self, variant, expected_variant, expected_fds):
        """Acquire Unix file descriptors in the variant."""
        variant, fd_list = acquire_fds(variant)

        # Check the list of acquired fds.
        if expected_fds is None:
            self.assertIsNone(fd_list)
        else:
            self.assertIsInstance(fd_list, Gio.UnixFDList)
            self.assertEqual(fd_list.peek_fds(), expected_fds)

        # Check the variant without fds.
        if expected_variant is None:
            self.assertIsNone(variant)
        else:
            self.assertIsInstance(variant, Variant)
            self.assertEqual(variant.unpack(), expected_variant.unpack())
            self.assertTrue(variant.equal(expected_variant))

        return variant, fd_list

    def _restore_fds(self, variant, fd_list, expected_variant):
        """Restore Unix file descriptors in the variant."""
        variant = restore_fds(variant, fd_list)

        # Check the variant with fds.
        if expected_variant is None:
            self.assertIsNone(variant)
        else:
            self.assertIsInstance(variant, Variant)
            self.assertEqual(variant.unpack(), expected_variant.unpack())
            self.assertTrue(variant.equal(expected_variant))

    def test_empty_fd_list(self):
        """Restore a value if the fd list is empty."""
        self._restore_fds(
            variant=get_variant(Int, 0),
            fd_list=Gio.UnixFDList(),
            expected_variant=get_variant(Int, 0)
        )

    def test_invalid_index(self):
        """Restore a value with an invalid index."""
        self._restore_fds(
            variant=get_variant(UnixFD, 2),
            fd_list=Gio.UnixFDList.new_from_array([self._r, self._w]),
            expected_variant=get_variant(UnixFD, -1)
        )

    def test_values_without_fds(self):
        """Swap values without fds."""
        self._swap_fds(
            in_variant=None,
            out_variant=None,
        )
        self._swap_fds(
            in_variant=get_variant(Int, 0),
            out_variant=get_variant(Int, 0),
        )
        self._swap_fds(
            in_variant=get_variant(Str, "Hi!"),
            out_variant=get_variant(Str, "Hi!"),
        )
        self._swap_fds(
            in_variant=get_variant(Tuple[Double], (1.0, )),
            out_variant=get_variant(Tuple[Double], (1.0, )),
        )
        self._swap_fds(
            in_variant=get_variant(List[Bool], [False]),
            out_variant=get_variant(List[Bool], [False]),
        )

    def test_value_with_fds(self):
        """Swap a basic value with fds."""
        self._swap_fds(
            in_variant=get_variant(UnixFD, self._r),
            out_variant=get_variant(UnixFD, 0),
            fds=[self._r],
        )

    def test_variant_with_fds(self):
        """Swap a variant with fds."""
        self._swap_fds(
            in_variant=get_variant(Variant, get_variant(
                UnixFD, self._r
            )),
            out_variant=get_variant(Variant, get_variant(
                UnixFD, 0
            )),
            fds=[self._r],
        )

    def test_tuple_with_fds(self):
        """Swap a tuple with fds."""
        self._swap_fds(
            in_variant=get_variant(Tuple[UnixFD, UnixFD], (
                self._r, self._w
            )),
            out_variant=get_variant(Tuple[UnixFD, UnixFD], (
                0, 1
            )),
            fds=[self._r, self._w],
        )

    def test_variant_tuple_with_fds(self):
        """Swap a tuple of variants with fds."""
        self._swap_fds(
            in_variant=get_variant(Tuple[Variant, Variant], (
                get_variant(UnixFD, self._r),
                get_variant(UnixFD, self._w),
            )),
            out_variant=get_variant(Tuple[Variant, Variant], (
                get_variant(UnixFD, 0),
                get_variant(UnixFD, 1),
            )),
            fds=[self._r, self._w],
        )

    def test_list_with_fds(self):
        """Swap a list with fds."""
        self._swap_fds(
            in_variant=get_variant(List[UnixFD], [
                self._r, self._w
            ]),
            out_variant=get_variant(List[UnixFD], [
                0, 1
            ]),
            fds=[self._r, self._w],
        )

    def test_variant_list_with_fds(self):
        """Swap a list of variants with fds."""
        self._swap_fds(
            in_variant=get_variant(List[Variant], [
                get_variant(UnixFD, self._r),
                get_variant(UnixFD, self._w)
            ]),
            out_variant=get_variant(List[Variant], [
                get_variant(UnixFD, 0),
                get_variant(UnixFD, 1)
            ]),
            fds=[self._r, self._w],
        )

    def test_dictionary_with_fds_values(self):
        """Swap a dictionary with fds values."""
        self._swap_fds(
            in_variant=get_variant(Dict[Str, UnixFD], {
                "r": self._r, "w": self._w
            }),
            out_variant=get_variant(Dict[Str, UnixFD], {
                "r": 0, "w": 1
            }),
            fds=[self._r, self._w]
        )

    def test_dictionary_with_fds_keys(self):
        """Swap a dictionary with fds keys."""
        self._swap_fds(
            in_variant=get_variant(Dict[UnixFD, Str], {
                self._r: "r", self._w: "w"
            }),
            out_variant=get_variant(Dict[UnixFD, Str], {
                0: "r", 1: "w"
            }),
            fds=[self._r, self._w]
        )

    def test_variant_dictionary_with_fds(self):
        """Swap a dictionary of variants with fds."""
        self._swap_fds(
            in_variant=get_variant(Dict[Str, Variant], {
                "r": get_variant(UnixFD, self._r),
                "w": get_variant(UnixFD, self._w)
            }),
            out_variant=get_variant(Dict[Str, Variant], {
                "r": get_variant(UnixFD, 0),
                "w": get_variant(UnixFD, 1)
            }),
            fds=[self._r, self._w]
        )

    def test_nested_variants_with_fds(self):
        """Swap nested variants with fds."""
        self._swap_fds(
            in_variant=get_variant(Variant, get_variant(
                Variant, get_variant(UnixFD, self._r)),
            ),
            out_variant=get_variant(Variant, get_variant(
                Variant, get_variant(UnixFD, 0)),
            ),
            fds=[self._r],
        )

    def test_nested_containers_with_fds(self):
        """Swap nested containers with fds."""
        self._swap_fds(
            in_variant=get_variant(
                Tuple[List[UnixFD], Dict[Str, Variant]],
                ([self._r], {"w": get_variant(UnixFD, self._w)})
            ),
            out_variant=get_variant(
                Tuple[List[UnixFD], Dict[Str, Variant]],
                ([0], {"w": get_variant(UnixFD, 1)})
            ),
            fds=[self._r, self._w],
        )


class DBusUnixCompatibilityTestCase(DBusExampleTestCase):
    """Test the Unix support compatibility with a real DBus connection."""

    def _get_message_bus(self, bus_address):
        """Get a message bus."""
        return UnixMessageBus(bus_address, error_mapper)


class DBusUnixExampleTestCase(DBusSpawnedTestCase):
    """Test Unix support with a real DBus connection."""

    def _get_service(self):
        """Get the example service."""
        return UnixExampleInterface()

    @classmethod
    def _get_message_bus(cls, bus_address):
        """Get a message bus."""
        return UnixMessageBus(bus_address)

    @classmethod
    def _get_proxy(cls, bus_address, **proxy_args):
        """Get a proxy of the example service."""
        message_bus = cls._get_message_bus(bus_address)
        return cls._get_service_proxy(message_bus, **proxy_args)

    def test_xml_specification(self):
        """Test the generated specification."""
        expected_xml = '''
        <node>
          <!--Specifies UnixExampleInterface-->
          <interface name="my.testing.UnixExample">
            <method name="Goodbye">
              <arg direction="in" name="name" type="s"></arg>
              <arg direction="out" name="return" type="h"></arg>
            </method>
            <method name="Hello">
              <arg direction="in" name="name_fd" type="h"></arg>
              <arg direction="out" name="return" type="s"></arg>
            </method>
            <property access="readwrite" name="Pipes" type="ah"></property>
            <signal name="Signal">
                <arg direction="out" name="name" type="s"></arg>
                <arg direction="out" name="name_fd" type="h"></arg>
            </signal>
            <method name="Trigger">
                <arg direction="in" name="name" type="s"></arg>
                <arg direction="in" name="name_fd" type="h"></arg>
            </method>
          </interface>
        </node>
        '''
        generated_xml = self.service.__dbus_xml__

        self.assertEqual(
            XMLGenerator.prettify_xml(expected_xml),
            XMLGenerator.prettify_xml(generated_xml)
        )

    def test_sync_calls(self):
        """Test DBus sync calls with fds."""
        self._add_client(self._call_hello_sync)
        self._add_client(self._call_goodbye_sync)
        self._run_test()

    @classmethod
    def _call_hello_sync(cls, bus_address):
        """Say sync hello to Foo."""
        proxy = cls._get_proxy(bus_address)
        fd = write_string("Foo")
        greeting = proxy.Hello(fd)
        assert greeting == "Hello, Foo!", greeting

    @classmethod
    def _call_goodbye_sync(cls, bus_address):
        """Say sync goodbye to Bar."""
        proxy = cls._get_proxy(bus_address)
        fd = proxy.Goodbye("Bar")
        greeting = read_string(fd)
        assert greeting == "Goodbye, Bar!", greeting

    def test_async_calls(self):
        """Test DBus async calls with fds."""
        self._add_client(self._call_hello_async)
        self._add_client(self._call_goodbye_async)
        self._run_test()

    @classmethod
    def _call_hello_async(cls, bus_address):
        """Say async hello to Foo."""
        proxy = cls._get_proxy(bus_address)

        @mocked
        def callback(call):
            greeting = call()
            assert greeting == "Hello, Foo!", greeting

        fd = write_string("Foo")
        proxy.Hello(fd, callback=callback)
        run_loop()

        callback.assert_called_once()

    @classmethod
    def _call_goodbye_async(cls, bus_address):
        """Say async goodbye to Bar."""
        proxy = cls._get_proxy(bus_address)

        @mocked
        def callback(call):
            greeting = read_string(call())
            assert greeting == "Goodbye, Bar!", greeting

        proxy.Goodbye("Bar", callback=callback)
        run_loop()

        callback.assert_called_once()

    def test_properties(self):
        """Test DBus properties with fds."""
        self._add_client(self._set_pipes)
        self._add_client(self._get_pipes)
        self._run_test()

    @classmethod
    def _set_pipes(cls, bus_address):
        proxy = cls._get_proxy(bus_address)
        pipes = list(map(write_string, ["1", "2", "3"]))
        proxy.Pipes = pipes

    @classmethod
    def _get_pipes(cls, bus_address):
        proxy = cls._get_proxy(bus_address)
        pipes = []

        while not pipes:
            pipes = proxy.Pipes

        values = list(map(read_string, pipes))
        assert values == ["1", "2", "3"]

    def test_signals(self):
        """Test DBus signals with fds."""
        event = self.context.Event()
        self._add_client(self._trigger_signal, event)
        self._add_client(self._watch_signal, event)
        self._run_test()

    @classmethod
    def _trigger_signal(cls, bus_address, event):
        event.wait()
        proxy = cls._get_proxy(bus_address)
        proxy.Trigger("Foo", write_string("Foo"))

    @classmethod
    def _watch_signal(cls, bus_address, event):
        proxy = cls._get_proxy(bus_address)

        @mocked
        def callback(name, name_fd):
            # GLib doesn't support fds in signals, so we
            # are not able to restore fds in this case.
            # Because of that, name_fd is just an index.
            assert name == "Foo"
            assert name_fd == 0

        proxy.Signal.connect(callback)
        event.set()
        run_loop()

        callback.assert_called_once()



#
# Say hello to the world.
# Start the server and run the client.
#
from common import HELLO_WORLD

if __name__ == "__main__":
    # Create a proxy of the object /org/example/HelloWorld
    # provided by the service org.example.HelloWorld
    proxy = HELLO_WORLD.get_proxy()

    # Call the DBus method Hello and print the return value.
    print(proxy.Hello("World"))



#
# Run the service org.example.HelloWorld.
#
from dasbus.loop import EventLoop
from dasbus.server.interface import dbus_interface
from dasbus.typing import Str
from common import HELLO_WORLD, SESSION_BUS
from dasbus.xml import XMLGenerator


@dbus_interface(HELLO_WORLD.interface_name)
class HelloWorld(object):
    """The DBus interface for HelloWorld."""

    def Hello(self, name: Str) -> Str:
        """Generate a greeting.

        :param name: someone to say hello
        :return: a greeting
        """
        return "Hello {}!".format(name)


if __name__ == "__main__":
    # Print the generated XML specification.
    print(XMLGenerator.prettify_xml(HelloWorld.__dbus_xml__))

    try:
        # Create an instance of the class HelloWorld.
        hello_world = HelloWorld()

        # Publish the instance at /org/example/HelloWorld.
        SESSION_BUS.publish_object(HELLO_WORLD.object_path, hello_world)

        # Register the service name org.example.HelloWorld.
        SESSION_BUS.register_service(HELLO_WORLD.service_name)

        # Start the event loop.
        loop = EventLoop()
        loop.run()
    finally:
        # Unregister the DBus service and objects.
        SESSION_BUS.disconnect()



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from unittest.mock import Mock

from dasbus.identifier import DBusInterfaceIdentifier, DBusObjectIdentifier, \
    DBusServiceIdentifier, DBusBaseIdentifier


class DBusIdentifierTestCase(unittest.TestCase):
    """Test DBus identifiers."""

    def assert_namespace(self, obj, namespace):
        """Check the DBus namespace object."""
        self.assertEqual(obj.namespace, namespace)

    def assert_interface(self, obj, interface_name):
        """Check the DBus interface object."""
        self.assertEqual(obj.interface_name, interface_name)

    def test_identifier(self):
        """Test the DBus identifier object."""
        identifier = DBusBaseIdentifier(
            namespace=("a", "b", "c")
        )
        self.assert_namespace(identifier, ("a", "b", "c"))

        identifier = DBusBaseIdentifier(
            basename="d",
            namespace=("a", "b", "c")
        )
        self.assert_namespace(identifier, ("a", "b", "c", "d"))

    def test_interface(self):
        """Test the DBus interface object."""
        interface = DBusInterfaceIdentifier(
            namespace=("a", "b", "c")
        )
        self.assert_namespace(interface, ("a", "b", "c"))
        self.assert_interface(interface, "a.b.c")

        interface = DBusInterfaceIdentifier(
            namespace=("a", "b", "c"),
            interface_version=1
        )
        self.assert_namespace(interface, ("a", "b", "c"))
        self.assert_interface(interface, "a.b.c1")

        interface = DBusInterfaceIdentifier(
            basename="d",
            namespace=("a", "b", "c"),
            interface_version=1
        )
        self.assert_namespace(interface, ("a", "b", "c", "d"))
        self.assert_interface(interface, "a.b.c.d1")

    def assert_object(self, obj, object_path):
        """Check the DBus object."""
        self.assertEqual(obj.object_path, object_path)

    def test_object(self):
        """Test the DBus object."""
        obj = DBusObjectIdentifier(
            namespace=("a", "b", "c")
        )
        self.assert_namespace(obj, ("a", "b", "c"))
        self.assert_interface(obj, "a.b.c")
        self.assert_object(obj, "/a/b/c")

        obj = DBusObjectIdentifier(
            namespace=("a", "b", "c"),
            object_version=2,
            interface_version=4
        )
        self.assert_namespace(obj, ("a", "b", "c"))
        self.assert_interface(obj, "a.b.c4")
        self.assert_object(obj, "/a/b/c2")

        obj = DBusObjectIdentifier(
            basename="d",
            namespace=("a", "b", "c"),
            object_version=2,
            interface_version=4
        )
        self.assert_namespace(obj, ("a", "b", "c", "d"))
        self.assert_interface(obj, "a.b.c.d4")
        self.assert_object(obj, "/a/b/c/d2")

    def assert_bus(self, obj, message_bus):
        """Check the DBus service object."""
        self.assertEqual(obj.message_bus, message_bus)

    def assert_service(self, obj, service_name):
        """Check the DBus service object."""
        self.assertEqual(obj.service_name, service_name)

    def test_service(self):
        """Test the DBus service object."""
        bus = Mock()
        service = DBusServiceIdentifier(
            namespace=("a", "b", "c"),
            message_bus=bus
        )
        self.assert_namespace(service, ("a", "b", "c"))
        self.assert_interface(service, "a.b.c")
        self.assert_object(service, "/a/b/c")
        self.assert_service(service, "a.b.c")
        self.assert_bus(service, bus)

        service = DBusServiceIdentifier(
            namespace=("a", "b", "c"),
            service_version=3,
            interface_version=5,
            object_version=7,
            message_bus=bus
        )
        self.assert_namespace(service, ("a", "b", "c"))
        self.assert_interface(service, "a.b.c5")
        self.assert_object(service, "/a/b/c7")
        self.assert_service(service, "a.b.c3")
        self.assert_bus(service, bus)

        service = DBusServiceIdentifier(
            basename="d",
            namespace=("a", "b", "c"),
            service_version=3,
            interface_version=5,
            object_version=7,
            message_bus=bus
        )
        self.assert_namespace(service, ("a", "b", "c", "d"))
        self.assert_interface(service, "a.b.c.d5")
        self.assert_object(service, "/a/b/c/d7")
        self.assert_service(service, "a.b.c.d3")
        self.assert_bus(service, bus)


class DBusServiceIdentifierTestCase(unittest.TestCase):
    """Test DBus service identifiers."""

    def test_get_proxy(self):
        """Test getting a proxy."""
        bus = Mock()
        namespace = ("a", "b", "c")

        service = DBusServiceIdentifier(
            namespace=namespace,
            message_bus=bus
        )

        obj = DBusObjectIdentifier(
            basename="object",
            namespace=namespace
        )

        service.get_proxy()
        bus.get_proxy.assert_called_with(
            "a.b.c",
            "/a/b/c",
            None
        )
        bus.reset_mock()

        service.get_proxy("/a/b/c/object")
        bus.get_proxy.assert_called_with(
            "a.b.c",
            "/a/b/c/object",
            None
        )
        bus.reset_mock()

        service.get_proxy(obj)
        bus.get_proxy.assert_called_with(
            "a.b.c",
            "/a/b/c/object",
            None
        )
        bus.reset_mock()

    def test_get_proxy_for_interface(self):
        """Test getting a proxy for an interface."""
        bus = Mock()
        namespace = ("a", "b", "c")

        service = DBusServiceIdentifier(
            namespace=namespace,
            message_bus=bus
        )

        interface = DBusInterfaceIdentifier(
            basename="interface",
            namespace=namespace
        )

        service.get_proxy(
            interface_name="a.b.c.interface"
        )
        bus.get_proxy.assert_called_with(
            "a.b.c",
            "/a/b/c",
            "a.b.c.interface"
        )
        bus.reset_mock()

        service.get_proxy(
            interface_name=interface
        )
        bus.get_proxy.assert_called_with(
            "a.b.c",
            "/a/b/c",
            "a.b.c.interface"
        )
        bus.reset_mock()

    def test_get_proxy_with_bus_arguments(self):
        """Test getting a proxy with an additional arguments."""
        bus = Mock()
        error_mapper = Mock()
        namespace = ("a", "b", "c")

        service = DBusServiceIdentifier(
            namespace=namespace,
            message_bus=bus
        )

        service.get_proxy(
            error_mapper=error_mapper
        )
        bus.get_proxy.assert_called_with(
            "a.b.c",
            "/a/b/c",
            None,
            error_mapper=error_mapper
        )
        bus.reset_mock()

        service.get_proxy(
            interface_name=service,
            error_mapper=error_mapper
        )
        bus.get_proxy.assert_called_with(
            "a.b.c",
            "/a/b/c",
            "a.b.c",
            error_mapper=error_mapper
        )
        bus.reset_mock()



#
# Client support for DBus objects
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from abc import ABCMeta, abstractmethod
from functools import partial

from dasbus.client.property import PropertyProxy
from dasbus.error import ErrorMapper
from dasbus.signal import Signal
from dasbus.constants import DBUS_FLAG_NONE
from dasbus.specification import DBusSpecification
from dasbus.typing import get_variant, get_variant_type, unwrap_variant

import gi
gi.require_version("Gio", "2.0")
gi.require_version("GLib", "2.0")
from gi.repository import Gio, GLib

__all__ = [
    "GLibClient",
    "AbstractClientObjectHandler",
    "ClientObjectHandler"
]


class GLibClient(object):
    """The low-level DBus client library based on GLib."""

    # Infinite timeout of a DBus call
    DBUS_TIMEOUT_NONE = GLib.MAXINT

    @classmethod
    def sync_call(cls, connection, service_name, object_path, interface_name,
                  method_name, parameters, reply_type, flags=DBUS_FLAG_NONE,
                  timeout=DBUS_TIMEOUT_NONE):
        """Synchronously call a DBus method.

        :return: a result of the DBus call
        """
        return connection.call_sync(
            service_name,
            object_path,
            interface_name,
            method_name,
            parameters,
            reply_type,
            flags,
            timeout,
            None
        )

    @classmethod
    def async_call(cls, connection, service_name, object_path, interface_name,
                   method_name, parameters, reply_type, callback,
                   callback_args=(), flags=DBUS_FLAG_NONE,
                   timeout=DBUS_TIMEOUT_NONE):
        """Asynchronously call a DBus method."""
        connection.call(
            service_name,
            object_path,
            interface_name,
            method_name,
            parameters,
            reply_type,
            flags,
            timeout,
            callback=cls._async_call_finish,
            user_data=(callback, callback_args)
        )

    @classmethod
    def _async_call_finish(cls, source_object, result_object, user_data):
        """Finish an asynchronous DBus method call."""
        # Prepare the user's callback.
        callback, callback_args = user_data

        # Call user's callback.
        callback(
            lambda: source_object.call_finish(result_object),
            *callback_args
        )

    @classmethod
    def subscribe_signal(cls, connection, service_name, object_path,
                         interface_name, signal_name, callback,
                         callback_args=(), flags=DBUS_FLAG_NONE):
        """Subscribe to a signal.

        :return: a callback to unsubscribe
        """
        subscription_id = connection.signal_subscribe(
            service_name,
            interface_name,
            signal_name,
            object_path,
            None,
            flags,
            callback=cls._signal_callback,
            user_data=(callback, callback_args)
        )

        return partial(
            cls._unsubscribe_signal,
            connection,
            subscription_id
        )

    @classmethod
    def _signal_callback(cls, connection, sender_name, object_path,
                         interface_name, signal_name, parameters, user_data):
        """A callback that is called when a DBus signal is emitted."""
        # Prepare the user's callback.
        callback, callback_args = user_data

        # Call user's callback.
        callback(parameters, *callback_args)

    @classmethod
    def _unsubscribe_signal(cls, connection, subscription_id):
        """Unsubscribe from a signal."""
        connection.signal_unsubscribe(subscription_id)

    @classmethod
    def is_timeout_error(cls, error):
        """Is it a timeout error?"""
        return isinstance(error, GLib.Error) \
            and error.matches(Gio.io_error_quark(), Gio.IOErrorEnum.TIMED_OUT)

    @classmethod
    def is_remote_error(cls, error):
        """Is it a remote DBus error?"""
        return isinstance(error, GLib.Error) \
            and Gio.DBusError.is_remote_error(error)

    @classmethod
    def get_remote_error_name(cls, error):
        """Get a DBus name of the remote DBus error."""
        return Gio.DBusError.get_remote_error(error)

    @classmethod
    def get_remote_error_message(cls, error):
        """Get a message of the remote DBus error."""
        name = cls.get_remote_error_name(error)
        message = error.message
        prefix = "{}:{}: ".format("GDBus.Error", name)

        if message.startswith(prefix):
            return message[len(prefix):]

        return message


class AbstractClientObjectHandler(metaclass=ABCMeta):
    """The abstract handler of a remote DBus object."""

    __slots__ = [
        "_message_bus",
        "_service_name",
        "_object_path",
        "_specification"
    ]

    def __init__(self, message_bus, service_name, object_path):
        """Create a new handler.

        :param message_bus: a message bus
        :param service_name: a DBus name of the service
        :param object_path: a DBus path the object
        """
        self._message_bus = message_bus
        self._service_name = service_name
        self._object_path = object_path
        self._specification = None

    @property
    def service_name(self):
        """DBus service name.

        :return: a DBus name
        """
        return self._service_name

    @property
    def object_path(self):
        """DBus object path.

        :return: a DBus path
        """
        return self._object_path

    @property
    def specification(self):
        """DBus specification."""
        if not self._specification:
            self._specification = self._get_specification()

        return self._specification

    @abstractmethod
    def _get_specification(self):
        """Introspect the DBus object.

        :return: a DBus specification
        """
        return DBusSpecification()

    def create_member(self, interface_name, member_name):
        """Create a member of the DBus object.

        :param interface_name: a name of the interface
        :param member_name: a name of the member
        :return: a signal, a method or a property
        """
        spec = self._find_member_spec(interface_name, member_name)
        handler = self._find_handler(type(spec))
        return handler(spec)

    def _find_member_spec(self, interface_name, member_name):
        """Find a specification of the DBus member.

        :param interface_name: a name of the interface
        :param member_name: a name of the member
        :return: a specification of the member
        """
        return self.specification.get_member(
            interface_name, member_name
        )

    def _find_handler(self, member_type):
        """Find a handler for the given member type.

        :param member_type: a type of the member
        :return: a callback
        """
        if member_type is DBusSpecification.Property:
            return self._get_property

        if member_type is DBusSpecification.Method:
            return self._get_method

        if member_type is DBusSpecification.Signal:
            return self._get_signal

        raise TypeError(
            "Unsupported type: {}".format(member_type.__name__)
        )

    @abstractmethod
    def _get_property(self, property_spec):
        """Get a proxy of the DBus property.

        :param property_spec: a property_specification
        :return: a property object
        """
        pass

    @abstractmethod
    def _get_method(self, method_spec):
        """Get a proxy of the DBus method.

        :param method_spec: a method specification
        :return: a callable object
        """
        pass

    @abstractmethod
    def _get_signal(self, signal_spec):
        """Get a proxy of the DBus signal.

        :param signal_spec: a signal specification
        :return: a signal object
        """
        pass

    @abstractmethod
    def disconnect_members(self):
        """Disconnect members of the DBus object.

        Unsubscribe from DBus signals and disconnect all
        registered callbacks of the proxy signals.
        """
        pass


class ClientObjectHandler(AbstractClientObjectHandler):
    """The client handler of a DBus object."""

    __slots__ = [
        "_client",
        "_signal_factory",
        "_error_mapper",
        "_subscriptions"
    ]

    def __init__(self, message_bus, service_name, object_path,
                 error_mapper=None, client=GLibClient,
                 signal_factory=Signal):
        """Create a new handler.

        :param message_bus: a message bus
        :param service_name: a DBus name of the service
        :param object_path: a DBus path the object
        :param error_mapper: a DBus error mapper
        :param client: a DBus client library
        :param signal_factory: a signal factory
        """
        super().__init__(message_bus, service_name, object_path)
        self._client = client
        self._signal_factory = signal_factory
        self._error_mapper = error_mapper or ErrorMapper()
        self._subscriptions = []

    def _get_specification(self):
        """Introspect the DBus object."""
        xml = self._call_method(
            "org.freedesktop.DBus.Introspectable",
            "Introspect",
            None,
            "(s)"
        )

        return DBusSpecification.from_xml(xml)

    def _get_signal(self, signal_spec):
        """Get a proxy of the DBus signal."""
        # Create a signal.
        signal = self._signal_factory()

        # Subscribe to a DBus signal.
        unsubscribe = self._client.subscribe_signal(
            self._message_bus.connection,
            self._service_name,
            self._object_path,
            signal_spec.interface_name,
            signal_spec.name,
            callback=self._signal_callback,
            callback_args=(signal.emit,)
        )

        # Keep the subscriptions.
        self._subscriptions.append(unsubscribe)
        self._subscriptions.append(signal.disconnect)

        return signal

    def _signal_callback(self, parameters, callback):
        """A callback that is called when a DBus signal is emitted."""
        callback(*unwrap_variant(parameters))

    def _get_property(self, property_spec):
        """Get a proxy of the DBus property."""
        getter = None
        setter = None

        if property_spec.readable:
            getter = partial(self._get_property_value, property_spec)

        if property_spec.writable:
            setter = partial(self._set_property_value, property_spec)

        return PropertyProxy(getter, setter)

    def _get_property_value(self, property_spec):
        """Get a value of the DBus property."""
        variant = self._call_method(
            "org.freedesktop.DBus.Properties",
            "Get",
            "(ss)",
            "(v)",
            property_spec.interface_name,
            property_spec.name
        )
        return unwrap_variant(variant)

    def _set_property_value(self, property_spec, property_value):
        """Set a value of the DBus property."""
        return self._call_method(
            "org.freedesktop.DBus.Properties",
            "Set",
            "(ssv)",
            None,
            property_spec.interface_name,
            property_spec.name,
            get_variant(property_spec.type, property_value)
        )

    def _get_method(self, method_spec):
        """Get a callable proxy of the DBus method."""
        return partial(
            self._call_method,
            method_spec.interface_name,
            method_spec.name,
            method_spec.in_type,
            method_spec.out_type
        )

    def _call_method(self, interface_name, method_name, in_type,
                     out_type, *parameters, **kwargs):
        """Call a DBus method.

        :return: a result of the call or None
        """
        # Create variants.
        if not parameters:
            parameters = None

        if in_type is not None:
            parameters = get_variant(in_type, parameters)

        # Create variant types.
        reply_type = None

        if out_type is not None:
            reply_type = get_variant_type(out_type)

        # Collect arguments.
        args = (
            self._message_bus.connection,
            self._service_name,
            self._object_path,
            interface_name,
            method_name,
            parameters,
            reply_type,
        )

        # Get the callback.
        callback = kwargs.pop("callback", None)
        callback_args = kwargs.pop("callback_args", tuple())

        # Choose the type of invocation.
        if not callback:
            return self._get_method_reply(
                self._client.sync_call,
                *args,
                **kwargs,
            )
        else:
            return self._client.async_call(
                *args,
                **kwargs,
                callback=self._method_callback,
                callback_args=(callback, callback_args)
            )

    def _method_callback(self, getter, callback, callback_args):
        """A callback of an asynchronous DBus method call."""
        callback(
            lambda: self._get_method_reply(getter),
            *callback_args
        )

    def _get_method_reply(self, call, *args, **kwargs):
        """Get a result of a DBus call.

        :param call: a callback
        :param args: arguments of the callback
        :param kwargs: keyword arguments of the callback
        :return: a result of the callback
        :raise: an exception raised by the callback
        """
        try:
            result = call(*args, **kwargs)
            return self._handle_method_result(result)
        except Exception as error:  # pylint: disable=broad-except
            return self._handle_method_error(error)

    def _handle_method_error(self, error):
        """Handle an error of a DBus call.

        If the call returned a DBus error, it will be mapped
        to a Python exception based on the rules defined in
        the available error mapper. It should be a subclass
        of DBusError.

        :param error: an exception raised during the call
        :raise Exception: if the call unexpectedly failed
        :raise TimeoutError: if the DBus call timed out
        :raise DBusError: if the call returned a DBus error
        """
        if self._client.is_remote_error(error):
            # Handle a remote DBus error.
            name = self._client.get_remote_error_name(error)
            cls = self._error_mapper.get_exception_type(name)
            message = self._client.get_remote_error_message(error)

            # Create a new exception.
            exception = cls(message)
            exception.dbus_name = name

            # Raise a new instance of the exception class.
            raise exception from None

        if self._client.is_timeout_error(error):
            # Handle a timeout error.
            raise TimeoutError(
                "The DBus call timeout was reached."
            ) from None

        # Or re-raise the original error.
        raise error

    def _handle_method_result(self, result):
        """Handle a result of a DBus call.

        :param result: a variant tuple
        """
        # Unwrap a variant tuple.
        values = unwrap_variant(result)

        # Return None if there are no values.
        if not values:
            return None

        # Return one value.
        if len(values) == 1:
            return values[0]

        # Return multiple values.
        return values

    def disconnect_members(self):
        """Disconnect members of the DBus object."""
        while self._subscriptions:
            callback = self._subscriptions.pop()
            callback()



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from unittest.mock import Mock

from dasbus.client.proxy import ObjectProxy, get_object_path


class DBusProxyTestCase(unittest.TestCase):
    """Test support for object proxies."""

    def test_get_object_path(self):
        """Test get_object_path."""
        proxy = ObjectProxy(Mock(), "my.service", "/my/path")
        self.assertEqual(get_object_path(proxy), "/my/path")

        with self.assertRaises(TypeError) as cm:
            get_object_path(None)

        self.assertEqual(
            "Invalid type 'NoneType'.",
            str(cm.exception)
        )



#
# Show the current hostname.
#
from dasbus.connection import SystemMessageBus

if __name__ == "__main__":
    # Create a representation of a system bus connection.
    bus = SystemMessageBus()

    # Create a proxy of the object /org/freedesktop/hostname1
    # provided by the service org.freedesktop.hostname1.
    proxy = bus.get_proxy(
        "org.freedesktop.hostname1",
        "/org/freedesktop/hostname1"
    )

    # Print a value of the DBus property Hostname.
    print(proxy.Hostname)



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest

from dasbus.typing import Int, Str, List, Double, UnixFD, Tuple, Bool, Dict
from dasbus.server.interface import dbus_interface, dbus_class, dbus_signal, \
    get_xml, DBusSpecificationGenerator, DBusSpecificationError, \
    accepts_additional_arguments, returns_multiple_arguments
from dasbus.xml import XMLGenerator


class InterfaceGeneratorTestCase(unittest.TestCase):

    def setUp(self):
        self.generator = DBusSpecificationGenerator
        self.maxDiff = None

    def _compare(self, cls, expected_xml):
        """Compare cls specification with the given xml."""
        generated_xml = get_xml(cls)  # pylint: disable=no-member

        self.assertEqual(
            XMLGenerator.prettify_xml(generated_xml),
            XMLGenerator.prettify_xml(expected_xml)
        )

    def test_exportable(self):
        """Test if the given name should be exported."""
        self.assertTrue(self.generator._is_exportable("Name"))
        self.assertTrue(self.generator._is_exportable("ValidName"))
        self.assertTrue(self.generator._is_exportable("ValidName123"))
        self.assertTrue(self.generator._is_exportable("Valid123Name456"))
        self.assertTrue(self.generator._is_exportable("VALIDNAME123"))

        self.assertFalse(self.generator._is_exportable("name"))
        self.assertFalse(self.generator._is_exportable("_Name"))
        self.assertFalse(self.generator._is_exportable("__Name"))
        self.assertFalse(self.generator._is_exportable("invalid_name"))
        self.assertFalse(self.generator._is_exportable("invalid_name_123"))
        self.assertFalse(self.generator._is_exportable("Invalid_Name"))
        self.assertFalse(self.generator._is_exportable("invalidname"))
        self.assertFalse(self.generator._is_exportable("invalid123"))
        self.assertFalse(self.generator._is_exportable("invalidName"))
        self.assertFalse(self.generator._is_exportable("123InvalidName"))

    def test_invalid_member(self):
        """Test invalid members."""

        class InvalidMemberClass(object):
            Test1 = None

        with self.assertRaises(DBusSpecificationError) as cm:
            self.generator._generate_interface(
                InvalidMemberClass,
                {},
                "my.example.Interface"
            )

        self.assertEqual(
            "Unsupported definition of DBus member 'Test1'.",
            str(cm.exception)
        )

    def test_is_method(self):
        """Test if methods are DBus methods."""

        class IsMethodClass(object):

            def Test1(self, a: Int):
                pass  # pragma: no cover

            Test2 = None

            @property
            def Test3(self):
                return None  # pragma: no cover

            @dbus_signal
            def Test4(self):
                pass  # pragma: no cover

        self.assertTrue(self.generator._is_method(IsMethodClass.Test1))
        self.assertFalse(self.generator._is_method(IsMethodClass.Test2))
        self.assertFalse(self.generator._is_method(IsMethodClass.Test3))
        self.assertFalse(self.generator._is_method(IsMethodClass.Test4))

    def test_invalid_method(self):
        """Test invalid methods."""

        class InvalidMethodClass(object):

            def Test1(self, a):
                pass  # pragma: no cover

            def Test2(self, a=None):
                pass  # pragma: no cover

            def Test3(self, a, b):
                pass  # pragma: no cover

            def Test4(self, a: Int, b: Str, c):
                pass  # pragma: no cover

            def Test5(self, a: Int, b: Str, c) -> Int:
                pass  # pragma: no cover

            def Test6(self, *arg):
                pass  # pragma: no cover

            def Test7(self, **kwargs):
                pass  # pragma: no cover

            def Test8(self, a: Int, b: Double, *, c, d=None):
                pass  # pragma: no cover

        self._check_invalid_method(
            InvalidMethodClass.Test1,
            "Undefined type of parameter 'a'."
        )
        self._check_invalid_method(
            InvalidMethodClass.Test2,
            "Undefined type of parameter 'a'."
        )
        self._check_invalid_method(
            InvalidMethodClass.Test3,
            "Undefined type of parameter 'a'."
        )
        self._check_invalid_method(
            InvalidMethodClass.Test4,
            "Undefined type of parameter 'c'."
        )
        self._check_invalid_method(
            InvalidMethodClass.Test5,
            "Undefined type of parameter 'c'."
        )
        self._check_invalid_method(
            InvalidMethodClass.Test6,
            "Only positional or keyword arguments are allowed."
        )
        self._check_invalid_method(
            InvalidMethodClass.Test7,
            "Only positional or keyword arguments are allowed."
        )
        self._check_invalid_method(
            InvalidMethodClass.Test8,
            "Only positional or keyword arguments are allowed."
        )

    def _check_invalid_method(self, method, message):
        """Try to generate a specification for an invalid method."""
        with self.assertRaises(DBusSpecificationError) as cm:
            self.generator._generate_method(method, method.__name__)

        self.assertEqual(str(cm.exception), message)

    def test_method(self):
        """Test interface with a method."""

        @dbus_interface("my.example.Interface")
        class MethodClass(object):

            def Method1(self):
                pass  # pragma: no cover

            def Method2(self, x: Int):
                pass  # pragma: no cover

            def Method3(self) -> Int:
                pass  # pragma: no cover

            def Method4(self, x: List[Double], y: UnixFD) -> Tuple[Int, Bool]:
                return 0, True  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies MethodClass-->
            <interface name="my.example.Interface">
                <method name="Method1"/>
                <method name="Method2">
                    <arg direction="in" name="x" type="i"/>
                </method>
                <method name="Method3">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <method name="Method4">
                    <arg direction="in" name="x" type="ad"/>
                    <arg direction="in" name="y" type="h"/>
                    <arg direction="out" name="return" type="(ib)"/>
                </method>
            </interface>
        </node>
        '''

        self._compare(MethodClass, expected_xml)

    def test_is_property(self):
        """Test property detection."""

        class IsPropertyClass(object):

            @property
            def Property1(self) -> Int:
                return 1  # pragma: no cover

            def _get_property2(self):
                return 2  # pragma: no cover

            Property2 = property(fget=_get_property2)

            def Property3(self) -> Int:
                return 3  # pragma: no cover

            @dbus_signal
            def Property4(self):
                pass  # pragma: no cover

        self.assertTrue(self.generator._is_property(
            IsPropertyClass.Property1
        ))
        self.assertTrue(self.generator._is_property(
            IsPropertyClass.Property2
        ))
        self.assertFalse(self.generator._is_property(
            IsPropertyClass.Property3
        ))
        self.assertFalse(self.generator._is_property(
            IsPropertyClass.Property4
        ))

    def test_property(self):
        """Test the interface with a property."""

        @dbus_interface("Interface", namespace=("my", "example"))
        class ReadWritePropertyClass(object):

            def __init__(self):
                self._property = 0  # pragma: no cover

            @property
            def Property(self) -> Int:
                return self._property  # pragma: no cover

            @Property.setter
            def Property(self, value: Int):
                self._property = value  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies ReadWritePropertyClass-->
            <interface name="my.example.Interface">
                <property name="Property" type="i" access="readwrite" />
            </interface>
        </node>
        '''

        self._compare(ReadWritePropertyClass, expected_xml)

    def test_invalid_property(self):
        """Test the interface with an invalid property."""

        class InvalidPropertyClass(object):

            InvalidProperty = property()

            @property
            def NoHintProperty(self):
                return 1  # pragma: no cover

        with self.assertRaises(DBusSpecificationError) as cm:
            self.generator._generate_property(
                InvalidPropertyClass.InvalidProperty,
                "InvalidProperty"
            )

        self.assertEqual(
            "DBus property 'InvalidProperty' is not accessible.",
            str(cm.exception)
        )

        with self.assertRaises(DBusSpecificationError) as cm:
            self.generator._generate_property(
                InvalidPropertyClass.NoHintProperty,
                "NoHintProperty"
            )

        self.assertEqual(
            "Undefined type of DBus property 'NoHintProperty'.",
            str(cm.exception)
        )

    def test_property_readonly(self):
        """Test readonly property."""

        @dbus_interface("Interface")
        class ReadonlyPropertyClass(object):

            def __init__(self):
                self._property = 0  # pragma: no cover

            @property
            def Property(self) -> Int:
                return self._property  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies ReadonlyPropertyClass-->
            <interface name="Interface">
                <property name="Property" type="i" access="read" />
            </interface>
        </node>
        '''

        self._compare(ReadonlyPropertyClass, expected_xml)

    def test_property_writeonly(self):
        """Test writeonly property."""

        @dbus_interface("Interface")
        class WriteonlyPropertyClass(object):

            def __init__(self):
                self._property = 0  # pragma: no cover

            def set_property(self, x: Int):
                self._property = x  # pragma: no cover

            Property = property(fset=set_property)

        expected_xml = '''
        <node>
            <!--Specifies WriteonlyPropertyClass-->
            <interface name="Interface">
                <property name="Property" type="i" access="write" />
            </interface>
        </node>
        '''

        self._compare(WriteonlyPropertyClass, expected_xml)

    def test_is_signal(self):
        """Test signal detection."""

        class IsSignalClass(object):

            @dbus_signal
            def Signal1(self):
                pass  # pragma: no cover

            @dbus_signal
            def Signal2(self, x: Int, y: Double):
                pass  # pragma: no cover

            Signal3 = dbus_signal()

            def Signal4(self):
                pass  # pragma: no cover

            @property
            def Signal5(self):
                return None  # pragma: no cover

            Signal6 = None

        self.assertTrue(self.generator._is_signal(IsSignalClass.Signal1))
        self.assertTrue(self.generator._is_signal(IsSignalClass.Signal2))
        self.assertTrue(self.generator._is_signal(IsSignalClass.Signal3))
        self.assertFalse(self.generator._is_signal(IsSignalClass.Signal4))
        self.assertFalse(self.generator._is_signal(IsSignalClass.Signal5))
        self.assertFalse(self.generator._is_signal(IsSignalClass.Signal6))

    def test_simple_signal(self):
        """Test interface with a simple signal."""

        @dbus_interface("Interface")
        class SimpleSignalClass(object):
            SimpleSignal = dbus_signal()

        expected_xml = '''
        <node>
            <!--Specifies SimpleSignalClass-->
            <interface name="Interface">
                <signal name="SimpleSignal"/>
            </interface>
        </node>
        '''

        self._compare(SimpleSignalClass, expected_xml)

    def test_signal(self):
        """Test interface with signals."""

        @dbus_interface("Interface")
        class SignalClass(object):

            @dbus_signal
            def SomethingHappened(self):
                """Signal that something happened."""
                pass  # pragma: no cover

            @dbus_signal
            def SignalSomething(self, x: Int, y: Str):
                """
                Signal that something happened.

                :param x: Parameter x.
                :param y: Parameter y
                """
                pass  # pragma: no cover

            def _emit_signals(self):  # pragma: no cover
                self.SomethingHappened.emit()  # pylint: disable=no-member
                self.SignalSomething.emit(0, "Something!")

        expected_xml = '''
        <node>
            <!--Specifies SignalClass-->
            <interface name="Interface">
                <signal name="SignalSomething">
                    <arg direction="out" name="x" type="i"/>
                    <arg direction="out" name="y" type="s"/>
                </signal>
                <signal name="SomethingHappened"/>
            </interface>
        </node>
        '''

        self._compare(SignalClass, expected_xml)

    def test_invalid_signal(self):
        """Test interface with an invalid signal."""

        class InvalidSignalClass(object):

            @dbus_signal
            def Signal1(self, x):
                pass  # pragma: no cover

            @dbus_signal
            def Signal2(self) -> Int:
                return 1  # pragma: no cover

        with self.assertRaises(DBusSpecificationError) as cm:
            self.generator._generate_signal(
                InvalidSignalClass.Signal1,
                "Signal1"
            )

        self.assertEqual(
            "Undefined type of parameter 'x'.",
            str(cm.exception)
        )

        with self.assertRaises(DBusSpecificationError) as cm:
            self.generator._generate_signal(
                InvalidSignalClass.Signal2,
                "Signal2"
            )

        self.assertEqual(
            "Invalid return type of DBus signal 'Signal2'.",
            str(cm.exception)
        )

    def test_override_method(self):
        """Test interface with overridden methods."""

        @dbus_interface("A")
        class AClass(object):

            def MethodA1(self, x: Int) -> Double:
                return x + 1.0  # pragma: no cover

            def MethodA2(self) -> Bool:
                return False  # pragma: no cover

            def MethodA3(self, x: List[Int]):
                pass  # pragma: no cover

        class BClass(AClass):

            def MethodA1(self, x: Int) -> Double:
                return x + 2.0  # pragma: no cover

            def MethodB1(self) -> Str:
                return ""  # pragma: no cover

        @dbus_interface("C")
        class CClass(object):

            def MethodC1(self) -> Tuple[Int, Int]:
                return 1, 2  # pragma: no cover

            def MethodC2(self, x: Double):
                pass  # pragma: no cover

        @dbus_interface("D")
        class DClass(BClass, CClass):

            def MethodD1(self) -> Tuple[Int, Str]:
                return 0, ""  # pragma: no cover

            def MethodA3(self, x: List[Int]):
                pass  # pragma: no cover

            def MethodC2(self, x: Double):
                pass  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies DClass-->
            <interface name="A">
                <method name="MethodA1">
                    <arg direction="in" name="x" type="i"/>
                    <arg direction="out" name="return" type="d"/>
                </method>
                <method name="MethodA2">
                    <arg direction="out" name="return" type="b"/>
                </method>
                <method name="MethodA3">
                    <arg direction="in" name="x" type="ai"/>
                </method>
            </interface>
            <interface name="C">
                <method name="MethodC1">
                    <arg direction="out" name="return" type="(ii)"/>
                </method>
                <method name="MethodC2">
                    <arg direction="in" name="x" type="d"/>
                </method>
            </interface>
            <interface name="D">
                <method name="MethodB1">
                    <arg direction="out" name="return" type="s"/>
                </method>
                <method name="MethodD1">
                    <arg direction="out" name="return" type="(is)"/>
                </method>
            </interface>
        </node>
        '''

        self._compare(DClass, expected_xml)

    def test_complex(self):
        """Test complex example."""

        @dbus_interface("ComplexA")
        class ComplexClassA(object):

            def __init__(self):
                self.a = 1  # pragma: no cover

            @dbus_signal
            def SignalA(self, x: Int):
                pass  # pragma: no cover

            @property
            def PropertyA(self) -> Double:
                return self.a + 2.5  # pragma: no cover

            def MethodA(self) -> Int:
                return self.a  # pragma: no cover

            def _methodA(self):
                return None  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies ComplexClassA-->
            <interface name="ComplexA">
                <method name="MethodA">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <property access="read" name="PropertyA" type="d"/>
                <signal name="SignalA">
                   <arg direction="out" name="x" type="i"/>
                </signal>
            </interface>
        </node>
        '''

        self._compare(ComplexClassA, expected_xml)

        @dbus_interface("ComplexB")
        class ComplexClassB(ComplexClassA):

            def __init__(self):  # pragma: no cover
                super().__init__()
                self.b = 2.0

            @dbus_signal
            def SignalB(self, x: Bool, y: Double, z: Tuple[Int, Int]):
                pass  # pragma: no cover

            @property
            def PropertyB(self) -> Double:
                return self.b  # pragma: no cover

            def MethodA(self) -> Int:
                return int(self.b)  # pragma: no cover

            def MethodB(self, a: Str, b: List[Double], c: Int) -> Int:
                return int(self.b)  # pragma: no cover

            def _methodB(self, x: Bool):
                pass  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies ComplexClassB-->
           <interface name="ComplexA">
                <method name="MethodA">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <property access="read" name="PropertyA" type="d"/>
                <signal name="SignalA">
                   <arg direction="out" name="x" type="i"/>
                </signal>
            </interface>
            <interface name="ComplexB">
                <method name="MethodB">
                    <arg direction="in" name="a" type="s"/>
                    <arg direction="in" name="b" type="ad"/>
                    <arg direction="in" name="c" type="i"/>
                    <arg direction="out" name="return" type="i"/>
                </method>
                <property access="read" name="PropertyB" type="d"/>
                <signal name="SignalB">
                    <arg direction="out" name="x" type="b"/>
                    <arg direction="out" name="y" type="d"/>
                    <arg direction="out" name="z" type="(ii)"/>
                </signal>
            </interface>
        </node>
        '''

        self._compare(ComplexClassB, expected_xml)

        @dbus_class
        class ComplexClassC(ComplexClassB):

            def MethodB(self, a: Str, b: List[Double], c: Int) -> Int:
                return 1  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies ComplexClassC-->
           <interface name="ComplexA">
                <method name="MethodA">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <property access="read" name="PropertyA" type="d"/>
                <signal name="SignalA">
                   <arg direction="out" name="x" type="i"/>
                </signal>
            </interface>
            <interface name="ComplexB">
                <method name="MethodB">
                    <arg direction="in" name="a" type="s"/>
                    <arg direction="in" name="b" type="ad"/>
                    <arg direction="in" name="c" type="i"/>
                    <arg direction="out" name="return" type="i"/>
                </method>
                <property access="read" name="PropertyB" type="d"/>
                <signal name="SignalB">
                    <arg direction="out" name="x" type="b"/>
                    <arg direction="out" name="y" type="d"/>
                    <arg direction="out" name="z" type="(ii)"/>
                </signal>
            </interface>
        </node>
        '''

        self._compare(ComplexClassC, expected_xml)

    def test_standard_interfaces(self):
        """Test members of standard interfaces."""

        @dbus_interface("InterfaceWithoutStandard")
        class ClassWithStandard(object):

            @property
            def Property(self) -> Int:
                return 1  # pragma: no cover

            def Method(self):
                pass  # pragma: no cover

            def Ping(self):
                # This method shouldn't be part of the interface
                pass  # pragma: no cover

            @dbus_signal
            def PropertiesChanged(self, a, b, c):
                # This signal shouldn't be part of the interface
                pass  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies ClassWithStandard-->
           <interface name="InterfaceWithoutStandard">
                <method name="Method" />
                <property access="read" name="Property" type="i"/>
            </interface>
        </node>
        '''

        self._compare(ClassWithStandard, expected_xml)

    def test_additional_arguments(self):
        """Test interface methods with additional arguments."""

        @dbus_interface("my.example.Interface")
        class AdditionalArgumentsClass(object):

            @accepts_additional_arguments
            def Method1(self, **info):
                pass  # pragma: no cover

            @accepts_additional_arguments
            def Method2(self, x: Int, **info):
                pass  # pragma: no cover

            @accepts_additional_arguments
            def Method3(self, *, call_info):
                pass  # pragma: no cover

            @accepts_additional_arguments
            def Method4(self, x: Int, *, call_info):
                pass  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies AdditionalArgumentsClass-->
            <interface name="my.example.Interface">
                <method name="Method1"/>
                <method name="Method2">
                    <arg direction="in" name="x" type="i"/>
                </method>
                <method name="Method3"/>
                <method name="Method4">
                    <arg direction="in" name="x" type="i"/>
                </method>
            </interface>
        </node>
        '''

        self._compare(AdditionalArgumentsClass, expected_xml)

    def test_multiple_output_arguments(self):
        """Test interface methods with multiple output arguments."""

        @dbus_interface("my.example.Interface")
        class MultipleOutputArgumentsClass(object):

            @returns_multiple_arguments
            def Method1(self):
                pass  # pragma: no cover

            @returns_multiple_arguments
            def Method2(self) -> None:
                pass  # pragma: no cover

            @returns_multiple_arguments
            def Method3(self) -> Tuple[Int, Bool, Str]:
                pass  # pragma: no cover

            @returns_multiple_arguments
            def Method4(self) -> Tuple[Tuple[Int], List[Str]]:
                pass  # pragma: no cover

        expected_xml = '''
        <node>
            <!--Specifies MultipleOutputArgumentsClass-->
            <interface name="my.example.Interface">
                <method name="Method1"/>
                <method name="Method2"/>
                <method name="Method3">
                    <arg direction="out" name="return_0" type="i"/>
                    <arg direction="out" name="return_1" type="b"/>
                    <arg direction="out" name="return_2" type="s"/>
                </method>
                <method name="Method4">
                    <arg direction="out" name="return_0" type="(i)"/>
                    <arg direction="out" name="return_1" type="as"/>
                </method>
            </interface>
        </node>
        '''

        self._compare(MultipleOutputArgumentsClass, expected_xml)

    def test_invalid_multiple_output_arguments(self):
        """Test interface methods with invalid output arguments."""

        class InvalidOutputArgumentsClass(object):

            @returns_multiple_arguments
            def Method1(self) -> Int:
                pass  # pragma: no cover

            @returns_multiple_arguments
            def Method2(self) -> List[Bool]:
                pass  # pragma: no cover

            @returns_multiple_arguments
            def Method3(self) -> Dict[Str, Bool]:
                pass  # pragma: no cover

            @returns_multiple_arguments
            def Method4(self) -> Tuple[Int]:
                pass  # pragma: no cover

        self._check_invalid_method(
            InvalidOutputArgumentsClass.Method1,
            "Expected a tuple of multiple arguments."
        )

        self._check_invalid_method(
            InvalidOutputArgumentsClass.Method2,
            "Expected a tuple of multiple arguments."
        )

        self._check_invalid_method(
            InvalidOutputArgumentsClass.Method3,
            "Expected a tuple of multiple arguments."
        )

        self._check_invalid_method(
            InvalidOutputArgumentsClass.Method4,
            "Expected a tuple of more than one argument."
        )



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from unittest import mock
from unittest.mock import Mock

from dasbus.client.proxy import disconnect_proxy
from dasbus.connection import AddressedMessageBus
from dasbus.error import ErrorMapper, get_error_decorator
from dasbus.server.interface import dbus_interface, dbus_signal, \
    accepts_additional_arguments, returns_multiple_arguments
from dasbus.typing import get_variant, Str, Int, Dict, Variant, List, \
    Tuple, Bool
from dasbus.xml import XMLGenerator
from threading import Event

from tests.lib_dbus import DBusThreadedTestCase

# Define the error mapper and decorator.
error_mapper = ErrorMapper()
dbus_error = get_error_decorator(error_mapper)


@dbus_error("my.testing.Error")
class ExampleException(Exception):
    pass


@dbus_interface("my.testing.Example")
class ExampleInterface(object):

    def __init__(self):
        self._knocked = False
        self._names = []
        self._values = [0]
        self._secrets = []

    @property
    def Name(self) -> Str:
        return "My example"

    @property
    def Value(self) -> Int:
        return self._values[-1]

    @Value.setter
    def Value(self, value: Int):
        self._values.append(value)
        self.PropertiesChanged(
            "my.testing.Example",
            {"Value": get_variant(Int, value)},
            ["Name"]
        )

    def _set_secret(self, secret: Str):
        self._secrets.append(secret)

    Secret = property(fset=_set_secret)

    def Knock(self):
        self._knocked = True
        self.Knocked()

    def Hello(self, name: Str) -> Str:
        self._names.append(name)
        self.Visited(name)
        return "Hello, {0}!".format(name)

    @dbus_signal
    def Knocked(self):
        pass

    @dbus_signal
    def Visited(self, name: Str):
        pass

    def Raise(self, message: Str):
        raise ExampleException(message)

    @dbus_signal
    def PropertiesChanged(self, interface: Str, changed: Dict[Str, Variant],
                          invalid: List[Str]):
        pass

    @accepts_additional_arguments
    def GetInfo(self, arg: Str, *, call_info) -> Str:
        return "{0}: {1}".format(arg, call_info)

    @returns_multiple_arguments
    def ReturnArgs(self) -> Tuple[Int, Bool, Str]:
        return 0, False, "zero"


class DBusExampleTestCase(DBusThreadedTestCase):
    """Test DBus support with a real DBus connection."""

    def _get_service(self):
        """Get the example service."""
        return ExampleInterface()

    @classmethod
    def _get_message_bus(cls, bus_address):
        """Get a message bus."""
        return AddressedMessageBus(bus_address, error_mapper)

    def _get_proxy(self, **proxy_args):
        """Get a proxy of the example service."""
        return self._get_service_proxy(self.message_bus, **proxy_args)

    def test_message_bus(self):
        """Test the message bus."""
        self.assertTrue(self.message_bus.check_connection())
        self.assertEqual(self.message_bus.address, self.bus.get_bus_address())
        self.assertEqual(self.message_bus.proxy.Ping(), None)

    def test_xml_specification(self):
        """Test the generated specification."""
        expected_xml = '''
        <node>
          <!--Specifies ExampleInterface-->
          <interface name="my.testing.Example">
            <method name="GetInfo">
              <arg direction="in" name="arg" type="s"></arg>
              <arg direction="out" name="return" type="s"></arg>
            </method>
            <method name="Hello">
              <arg direction="in" name="name" type="s"></arg>
              <arg direction="out" name="return" type="s"></arg>
            </method>
            <method name="Knock"></method>
            <signal name="Knocked"></signal>
            <property access="read" name="Name" type="s"></property>
            <method name="Raise">
              <arg direction="in" name="message" type="s"></arg>
            </method>
            <method name="ReturnArgs">
              <arg direction="out" name="return_0" type="i"></arg>
              <arg direction="out" name="return_1" type="b"></arg>
              <arg direction="out" name="return_2" type="s"></arg>
            </method>
            <property access="write" name="Secret" type="s"></property>
            <property access="readwrite" name="Value" type="i"></property>
            <signal name="Visited">
              <arg direction="out" name="name" type="s"></arg>
            </signal>
          </interface>
        </node>
        '''

        generated_xml = self.service.__dbus_xml__

        self.assertEqual(
            XMLGenerator.prettify_xml(expected_xml),
            XMLGenerator.prettify_xml(generated_xml)
        )

    def test_knock(self):
        """Call a simple DBus method."""
        self.assertEqual(self.service._knocked, False)

        def test():
            proxy = self._get_proxy()
            self.assertEqual(None, proxy.Knock())

        self._add_client(test)
        self._run_test()

        self.assertEqual(self.service._knocked, True)

    def test_hello(self):
        """Call a DBus method."""

        def test1():
            proxy = self._get_proxy()
            self.assertEqual("Hello, Foo!", proxy.Hello("Foo"))

        def test2():
            proxy = self._get_proxy()
            self.assertEqual("Hello, Bar!", proxy.Hello("Bar"))

        self._add_client(test1)
        self._add_client(test2)
        self._run_test()

        self.assertEqual(sorted(self.service._names), ["Bar", "Foo"])

    def test_timeout(self):
        """Call a DBus method with a timeout."""

        def test1():
            proxy = self._get_proxy()
            proxy.Hello("Foo", timeout=1000)

        def test2():
            proxy = self._get_proxy()

            with self.assertRaises(TimeoutError):
                proxy.Hello("Bar", timeout=0)

        self._add_client(test1)
        self._add_client(test2)
        self._run_test()

        self.assertEqual(sorted(self.service._names), ["Bar", "Foo"])

    def test_name(self):
        """Use a DBus read-only property."""

        def test1():
            proxy = self._get_proxy()
            self.assertEqual("My example", proxy.Name)

        def test2():
            proxy = self._get_proxy()
            self.assertEqual("My example", proxy.Name)

        def test3():
            proxy = self._get_proxy()
            with self.assertRaises(AttributeError) as cm:
                proxy.Name = "Another example"

            self.assertEqual(
                "Can't set DBus property.",
                str(cm.exception)
            )

            self.assertEqual("My example", proxy.Name)

        self._add_client(test1)
        self._add_client(test2)
        self._add_client(test3)
        self._run_test()

    def test_secret(self):
        """Use a DBus write-only property."""

        def test1():
            proxy = self._get_proxy()
            proxy.Secret = "Secret 1"

        def test2():
            proxy = self._get_proxy()
            proxy.Secret = "Secret 2"

        def test3():
            proxy = self._get_proxy()
            with self.assertRaises(AttributeError) as cm:
                self.fail(proxy.Secret)

            self.assertEqual(
                "Can't read DBus property.",
                str(cm.exception)
            )

        self._add_client(test1)
        self._add_client(test2)
        self._add_client(test3)
        self._run_test()

        self.assertEqual(sorted(self.service._secrets), [
            "Secret 1",
            "Secret 2"
        ])

    def test_value(self):
        """Use a DBus read-write property."""

        def test1():
            proxy = self._get_proxy()
            self.assertIn(proxy.Value, (0, 3, 4))
            proxy.Value = 1
            self.assertIn(proxy.Value, (1, 3, 4))
            proxy.Value = 2
            self.assertIn(proxy.Value, (2, 3, 4))

        def test2():
            proxy = self._get_proxy()
            self.assertIn(proxy.Value, (0, 1, 2))
            proxy.Value = 3
            self.assertIn(proxy.Value, (3, 1, 2))
            proxy.Value = 4
            self.assertIn(proxy.Value, (4, 1, 2))

        self._add_client(test1)
        self._add_client(test2)
        self._run_test()

        self.assertEqual(sorted(self.service._values), [0, 1, 2, 3, 4])
        self.assertEqual(self.service._values[0], 0)
        self.assertLess(self.service._values.index(1),
                        self.service._values.index(2))
        self.assertLess(self.service._values.index(3),
                        self.service._values.index(4))

    def test_knocked(self):
        """Use a simple DBus signal."""
        event = Event()
        knocked = Mock()

        def callback():
            knocked("Knocked!")

        def test_1():
            proxy = self._get_proxy()
            proxy.Knocked.connect(callback)
            event.set()

        def test_2():
            event.wait()
            proxy = self._get_proxy()
            proxy.Knock()
            proxy.Knock()
            proxy.Knock()

        self._add_client(test_1)
        self._add_client(test_2)
        self._run_test()

        knocked.assert_has_calls([
            mock.call("Knocked!"),
            mock.call("Knocked!"),
            mock.call("Knocked!")
        ])

    def test_visited(self):
        """Use a DBus signal."""
        event = Event()
        visited = Mock()

        def callback(name):
            visited("Visited by {0}.".format(name))

        def test1():
            proxy = self._get_proxy()
            proxy.Visited.connect(callback)
            event.set()

        def test2():
            event.wait()
            proxy = self._get_proxy()
            proxy.Hello("Foo")
            proxy.Hello("Bar")

        self._add_client(test1)
        self._add_client(test2)
        self._run_test()

        visited.assert_has_calls([
            mock.call("Visited by Foo."),
            mock.call("Visited by Bar.")
        ])

    def test_unsubscribed(self):
        """Use an unsubscribed DBus signal."""
        event = Event()
        knocked = Mock()

        def callback():
            knocked("Knocked!")

        def test_1():
            proxy = self._get_proxy()
            proxy.Knocked.connect(callback)
            disconnect_proxy(proxy)
            event.set()

        def test_2():
            event.wait()
            proxy = self._get_proxy()
            proxy.Knock()
            proxy.Knock()
            proxy.Knock()

        self._add_client(test_1)
        self._add_client(test_2)
        self._run_test()

        knocked.assert_not_called()

    def test_asynchronous(self):
        """Call a DBus method asynchronously."""
        returned = Mock()

        def callback(call, number):
            returned(number, call())

        def test():
            proxy = self._get_proxy()
            proxy.Hello("Foo", callback=callback, callback_args=(1, ))
            proxy.Hello("Foo", callback=callback, callback_args=(2, ))
            proxy.Hello("Bar", callback=callback, callback_args=(3, ))

        self._add_client(test)
        self._run_test()

        returned.assert_has_calls([
            mock.call(1, "Hello, Foo!"),
            mock.call(2, "Hello, Foo!"),
            mock.call(3, "Hello, Bar!"),
        ])

    def test_error(self):
        """Handle a DBus error."""
        raised = Mock()

        def callback(call, number):
            try:
                call()
            except ExampleException as e:
                raised(number, str(e))

        def test1():
            proxy = self._get_proxy()
            proxy.Raise("Foo failed!", callback=callback, callback_args=(1, ))
            proxy.Raise("Foo failed!", callback=callback, callback_args=(2, ))
            proxy.Raise("Bar failed!", callback=callback, callback_args=(3, ))

        def test2():
            proxy = self._get_proxy()

            try:
                proxy.Raise("My message")
            except ExampleException as e:
                self.assertEqual(str(e), "My message")
            else:
                self.fail("Exception wasn't raised!")

        self._add_client(test1)
        self._add_client(test2)

        with self.assertLogs(level='WARN'):
            self._run_test()

        raised.assert_has_calls([
            mock.call(1, "Foo failed!"),
            mock.call(2, "Foo failed!"),
            mock.call(3, "Bar failed!"),
        ])

    def test_properties_changed(self):
        """Test the PropertiesChanged signal."""
        event = Event()
        callback = Mock()

        def test_1():
            proxy = self._get_proxy()
            proxy.PropertiesChanged.connect(callback)
            event.set()

        def test_2():
            event.wait()
            proxy = self._get_proxy()
            proxy.Value = 10

        self._add_client(test_1)
        self._add_client(test_2)
        self._run_test()

        callback.assert_called_once_with(
            "my.testing.Example",
            {"Value": get_variant(Int, 10)},
            ["Name"]
        )

    def test_interface(self):
        """Use a specific DBus interface."""

        def test_1():
            proxy = self._get_proxy(interface_name="my.testing.Example")
            proxy.Knock()

            with self.assertRaises(AttributeError):
                proxy.Ping()

        def test_2():
            proxy = self._get_proxy(interface_name="org.freedesktop.DBus.Peer")
            proxy.Ping()

            with self.assertRaises(AttributeError):
                proxy.Knock()

        self._add_client(test_1)
        self._add_client(test_2)
        self._run_test()

    def test_additional_arguments(self):
        """Call a DBus method."""

        def test1():
            proxy = self._get_proxy()
            self.assertEqual(
                proxy.GetInfo("Foo"),
                "Foo: {'sender': ':1.0'}"
            )

        def test2():
            proxy = self._get_proxy()
            self.assertEqual(
                proxy.GetInfo("Bar"),
                "Bar: {'sender': ':1.0'}"
            )

        self._add_client(test1)
        self._add_client(test2)
        self._run_test()

    def test_multiple_output_arguments(self):
        """Call a DBus method with multiple output arguments."""

        def test1():
            proxy = self._get_proxy()
            self.assertEqual(
                proxy.ReturnArgs(),
                (0, False, "zero")
            )

        self._add_client(test1)
        self._run_test()



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from unittest.mock import Mock

from dasbus.server.interface import dbus_signal
from dasbus.signal import Signal


class DBusSignalTestCase(unittest.TestCase):
    """Test DBus signals."""

    def test_create_signal(self):
        """Create a signal."""
        class Interface(object):

            @dbus_signal
            def Signal(self):
                pass

        interface = Interface()
        signal = interface.Signal
        self.assertIsInstance(signal, Signal)
        self.assertTrue(hasattr(interface, "__dbus_signal_signal"))
        self.assertEqual(getattr(interface, "__dbus_signal_signal"), signal)

    def test_emit_signal(self):
        """Emit a signal."""
        class Interface(object):

            @dbus_signal
            def Signal(self, a, b, c):
                pass

        interface = Interface()
        signal = interface.Signal

        callback = Mock()
        signal.connect(callback)  # pylint: disable=no-member

        signal.emit(1, 2, 3)  # pylint: disable=no-member
        callback.assert_called_once_with(1, 2, 3)
        callback.reset_mock()

        signal.emit(4, 5, 6)  # pylint: disable=no-member
        callback.assert_called_once_with(4, 5, 6)
        callback.reset_mock()

    def test_disconnect_signal(self):
        """Disconnect a signal."""
        class Interface(object):

            @dbus_signal
            def Signal(self):
                pass

        interface = Interface()
        callback = Mock()
        interface.Signal.connect(callback)  # pylint: disable=no-member

        interface.Signal()
        callback.assert_called_once_with()
        callback.reset_mock()

        interface.Signal.disconnect(callback)  # pylint: disable=no-member
        interface.Signal()
        callback.assert_not_called()

        interface.Signal.connect(callback)  # pylint: disable=no-member
        interface.Signal.disconnect()  # pylint: disable=no-member
        interface.Signal()
        callback.assert_not_called()

    def test_signals(self):
        """Test a class with two signals."""
        class Interface(object):

            @dbus_signal
            def Signal1(self):
                pass

            @dbus_signal
            def Signal2(self):
                pass

        interface = Interface()
        signal1 = interface.Signal1
        signal2 = interface.Signal2

        self.assertNotEqual(signal1, signal2)

        callback1 = Mock()
        signal1.connect(callback1)  # pylint: disable=no-member

        callback2 = Mock()
        signal2.connect(callback2)  # pylint: disable=no-member

        signal1.emit()  # pylint: disable=no-member
        callback1.assert_called_once_with()
        callback2.assert_not_called()
        callback1.reset_mock()

        signal2.emit()  # pylint: disable=no-member
        callback1.assert_not_called()
        callback2.assert_called_once_with()

    def test_instances(self):
        """Test two instances of the class with a signal."""
        class Interface(object):

            @dbus_signal
            def Signal(self):
                pass

        interface1 = Interface()
        signal1 = interface1.Signal

        interface2 = Interface()
        signal2 = interface2.Signal
        self.assertNotEqual(signal1, signal2)

        callback = Mock()
        signal1.connect(callback)  # pylint: disable=no-member

        callback2 = Mock()
        signal2.connect(callback2)  # pylint: disable=no-member

        signal1.emit()  # pylint: disable=no-member
        callback.assert_called_once_with()
        callback2.assert_not_called()
